#!/usr/bin/env python
"""
Modelica parse Tree to AST tree.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import copy
import hashlib
import logging
import os
import pickle
import platform
import sqlite3
import time
from collections import OrderedDict, deque
from datetime import timedelta
from pathlib import Path
from typing import Dict, List, Optional, Union  # noqa: F401

import antlr4
import antlr4.Parser

import pymoca

from . import ast
from .generated.ModelicaLexer import ModelicaLexer
from .generated.ModelicaListener import ModelicaListener
from .generated.ModelicaParser import ModelicaParser


# TODO
#  - Named function arguments (note that either all have to be named, or none)
#  - Make sure slice indices (eventually) evaluate to integers


logger = logging.getLogger("pymoca")


DEFAULT_MODEL_CACHE_DB = "model_txt_cache.db"


class ModelicaFile:
    def __init__(self, **kwargs):
        self.within = []  # type: List[ast.ComponentRef]
        self.classes = OrderedDict()  # type: OrderedDict[str, ast.Class]
        super().__init__(**kwargs)


class ASTListener(ModelicaListener):
    def __init__(self):
        self.file_node = None  # type: ModelicaFile
        self.ast = {}  # type: Dict[ast.Node]
        self.class_nodes = deque([ast.Class()])  # type: deque[ast.Class]
        self.comp_clause = None  # type: ast.ComponentClause
        self.eq_sect = None  # type: ast.EquationSection
        self.alg_sect = None  # type: ast.AlgorithmSection
        self.symbol_node = None  # type: ast.Symbol
        self.eq_comment = None  # type: str
        self.sym_count = 0  # type: int
        self.in_extends_clause = False

    @property
    def class_node(self):
        return self.class_nodes[-1]

    # FILE ===========================================================

    def enterStored_definition(self, ctx: ModelicaParser.Stored_definitionContext):
        file_node = ModelicaFile()
        self.ast[ctx] = file_node
        self.file_node = file_node

    def exitStored_definition(self, ctx: ModelicaParser.Stored_definitionContext):
        within = []
        if ctx.component_reference() is not None:
            within = [self.ast[ctx.component_reference()]]
        self.file_node.within = within

        for class_node in [self.ast[e] for e in ctx.stored_definition_class()]:
            self.ast[ctx].classes[class_node.name] = class_node
        self.file_node = self.ast[ctx]

    # CLASS ===========================================================

    def exitStored_definition_class(self, ctx: ModelicaParser.Stored_definition_classContext):
        class_node = self.ast[ctx.class_definition()]
        class_node.final = ctx.FINAL() is not None
        self.ast[ctx] = class_node

    def enterClass_definition(self, ctx: ModelicaParser.Class_definitionContext):
        class_node = ast.Class()
        class_node.encapsulated = ctx.ENCAPSULATED() is not None
        class_node.partial = ctx.class_prefixes().PARTIAL() is not None
        class_node.type = ctx.class_prefixes().class_type().getText()

        self.class_nodes.append(class_node)

        self.ast[ctx] = class_node

    def exitClass_definition(self, ctx: ModelicaParser.Class_definitionContext):
        class_node = self.class_nodes.pop()
        self.class_node.classes[class_node.name] = class_node

    def exitShort_class_definition(self, ctx):
        self.ast[ctx] = ast.ShortClassDefinition(
            name=ctx.IDENT().getText(),
            type=ctx.class_prefixes().class_type().getText(),
            component=self.ast[ctx.component_reference()],
        )

    def exitClass_spec_comp(self, ctx: ModelicaParser.Class_spec_compContext):
        class_node = self.class_node
        class_node.name = ctx.IDENT()[0].getText()
        class_node.comment = self.ast[ctx.string_comment()]

    def exitClass_spec_base(self, ctx: ModelicaParser.Class_spec_baseContext):
        class_node = self.class_node
        class_node.name = ctx.IDENT().getText()
        class_node.comment = self.ast[ctx.comment()]

        if ctx.class_modification() is not None:
            class_modification = self.ast[ctx.class_modification()]
        else:
            class_modification = ast.ClassModification()
        extends_clause = ast.ExtendsClause(
            component=self.ast[ctx.component_reference()], class_modification=class_modification
        )
        class_node.extends.append(extends_clause)

    def exitComposition(self, ctx: ModelicaParser.CompositionContext):
        for clause in self.ast[ctx.epriv]:
            if isinstance(clause, ast.ComponentClause):
                for symbol in clause.symbol_list:
                    symbol.visibility = ast.Visibility.PRIVATE
            elif isinstance(clause, ast.ExtendsClause):
                clause.visibility = ast.Visibility.PRIVATE

        if ctx.epub is not None:
            for clause in self.ast[ctx.epub]:
                if isinstance(clause, ast.ComponentClause):
                    for symbol in clause.symbol_list:
                        symbol.visibility = ast.Visibility.PUBLIC
                elif isinstance(clause, ast.ExtendsClause):
                    clause.visibility = ast.Visibility.PUBLIC

        if ctx.epro is not None:
            for clause in self.ast[ctx.epro]:
                if isinstance(clause, ast.ComponentClause):
                    for symbol in clause.symbol_list:
                        symbol.visibility = ast.Visibility.PROTECTED
                elif isinstance(clause, ast.ExtendsClause):
                    clause.visibility = ast.Visibility.PROTECTED

        for eqlist in [self.ast[e] for e in ctx.equation_section()]:
            if eqlist is not None:
                if eqlist.initial:
                    self.class_node.initial_equations += eqlist.equations
                else:
                    self.class_node.equations += eqlist.equations

        for alglist in [self.ast[e] for e in ctx.algorithm_section()]:
            if alglist is not None:
                if alglist.initial:
                    self.class_node.initial_statements += alglist.statements
                else:
                    self.class_node.statements += alglist.statements

        if ctx.comp_annotation is not None:
            self.class_node.annotation = self.ast[ctx.comp_annotation]

    def exitArgument(self, ctx: ModelicaParser.ArgumentContext):
        argument = ast.ClassModificationArgument()
        if ctx.element_modification_or_replaceable() is not None:
            argument.value = self.ast[ctx.element_modification_or_replaceable()]
            argument.redeclare = False
        else:
            argument.value = self.ast[ctx.element_redeclaration()]
            argument.redeclare = True
        self.ast[ctx] = argument

    def exitArgument_list(self, ctx: ModelicaParser.Argument_listContext):
        self.ast[ctx] = [self.ast[a] for a in ctx.argument()]

    def exitClass_modification(self, ctx: ModelicaParser.Class_modificationContext):
        arguments = []
        if ctx.argument_list() is not None:
            arguments = self.ast[ctx.argument_list()]
        self.ast[ctx] = ast.ClassModification(arguments=arguments)

    def enterEquation_section(self, ctx: ModelicaParser.Equation_sectionContext):
        eq_sect = ast.EquationSection(initial=ctx.INITIAL() is not None)
        self.ast[ctx] = eq_sect
        self.eq_sect = eq_sect

    def exitEquation_section(self, ctx: ModelicaParser.Equation_sectionContext):
        eq_sect = self.ast[ctx]
        eq_sect.equations.extend(self.ast[ctx.equation_block()])

    def exitEquation_block(self, ctx: ModelicaParser.Equation_blockContext):
        self.ast[ctx] = [self.ast[e] for e in ctx.equation()]

    def exitStatement_block(self, ctx):
        self.ast[ctx] = [self.ast[e] for e in ctx.statement()]

    def enterAlgorithm_section(self, ctx: ModelicaParser.Algorithm_sectionContext):
        alg_sect = ast.AlgorithmSection(initial=ctx.INITIAL() is not None)
        self.ast[ctx] = alg_sect
        self.alg_sect = alg_sect

    def exitAlgorithm_section(self, ctx: ModelicaParser.Algorithm_sectionContext):
        alg_sect = self.ast[ctx]
        alg_sect.statements.extend(self.ast[ctx.statement_block()])

    # EQUATION ===========================================================

    def enterEquation(self, ctx: ModelicaParser.EquationContext):
        pass

    def exitEquation(self, ctx):
        self.ast[ctx] = self.ast[ctx.equation_options()]
        try:
            self.ast[ctx].comment = self.ast[ctx.comment()]
        except AttributeError:
            pass

    def exitEquation_simple(self, ctx: ModelicaParser.Equation_simpleContext):
        self.ast[ctx] = ast.Equation(
            left=self.ast[ctx.simple_expression()],
            right=self.ast[ctx.expression()],
        )

    def exitEquation_if(self, ctx: ModelicaParser.Equation_ifContext):
        self.ast[ctx] = self.ast[ctx.if_equation()]

    def exitEquation_for(self, ctx: ModelicaParser.Equation_forContext):
        self.ast[ctx] = self.ast[ctx.for_equation()]

    def exitEquation_connect_clause(self, ctx: ModelicaParser.Equation_connect_clauseContext):
        self.ast[ctx] = self.ast[ctx.connect_clause()]

    def exitArgument_expression(self, ctx: ModelicaParser.Argument_expressionContext):
        self.ast[ctx] = self.ast[ctx.expression()]

    def exitIf_equation(self, ctx: ModelicaParser.If_equationContext):
        blocks = [self.ast[b] for b in ctx.blocks]
        conditions = [self.ast[c] for c in ctx.conditions]
        if len(conditions) == len(blocks) - 1:
            conditions.append(True)
        self.ast[ctx] = ast.IfEquation(conditions=conditions, blocks=blocks)

    def exitWhen_equation(self, ctx: ModelicaParser.When_equationContext):
        blocks = [self.ast[b] for b in ctx.blocks]
        conditions = [self.ast[c] for c in ctx.conditions]
        if len(conditions) == len(blocks) - 1:
            conditions.append(True)
        self.ast[ctx] = ast.WhenEquation(conditions=conditions, blocks=blocks)

    def exitFor_equation(self, ctx: ModelicaParser.For_equationContext):
        self.ast[ctx] = ast.ForEquation(
            indices=self.ast[ctx.for_indices()],
            equations=self.ast[ctx.equation_block()],
        )

    def exitConnect_clause(self, ctx: ModelicaParser.Connect_clauseContext):
        self.ast[ctx] = ast.ConnectClause(
            left=self.ast[ctx.component_reference()[0]],
            right=self.ast[ctx.component_reference()[1]],
        )

    # STATEMENT ==========================================================

    # TODO:
    # - Missing statement types:
    #   + Function calls (inside current function).
    #   + Break
    #   + Return
    #   + While
    #   + When? (also in equation missing)

    def enterStatement(self, ctx: ModelicaParser.StatementContext):
        pass

    def exitStatement(self, ctx: ModelicaParser.StatementContext):
        self.ast[ctx] = self.ast[ctx.statement_options()]
        try:
            self.ast[ctx].comment = self.ast[ctx.comment()]
        except AttributeError:
            pass

    def exitStatement_component_reference(
        self, ctx: ModelicaParser.Statement_component_referenceContext
    ):
        self.ast[ctx] = ast.AssignmentStatement(
            left=[self.ast[ctx.component_reference()]],
            right=self.ast[ctx.expression()],
        )

    def exitStatement_component_function(
        self, ctx: ModelicaParser.Statement_component_functionContext
    ):
        all_comp_refs = [self.ast[x] for x in ctx.component_reference()]

        right = ast.Expression(
            operator=all_comp_refs[-1],
            operands=[
                self.ast[x.expression()]
                for x in ctx.function_call_args().function_arguments().function_argument()
            ],
        )

        self.ast[ctx] = ast.AssignmentStatement(left=all_comp_refs[:-1], right=right)

    def exitStatement_if(self, ctx: ModelicaParser.Statement_ifContext):
        self.ast[ctx] = self.ast[ctx.if_statement()]

    def exitStatement_for(self, ctx: ModelicaParser.Statement_forContext):
        self.ast[ctx] = self.ast[ctx.for_statement()]

    def exitStatement_when(self, ctx: ModelicaParser.Equation_whenContext):
        self.ast[ctx] = self.ast[ctx.when_equation()]

    def exitIf_statement(self, ctx: ModelicaParser.If_statementContext):
        blocks = [self.ast[b] for b in ctx.blocks]
        conditions = [self.ast[c] for c in ctx.conditions]
        if len(conditions) == len(blocks) - 1:
            conditions.append(True)
        self.ast[ctx] = ast.IfStatement(conditions=conditions, blocks=blocks)

    def exitWhen_statement(self, ctx: ModelicaParser.When_statementContext):
        blocks = [self.ast[b] for b in ctx.blocks]
        conditions = [self.ast[c] for c in ctx.conditions]
        if len(conditions) == len(blocks) - 1:
            conditions.append(True)
        self.ast[ctx] = ast.WhenStatement(conditions=conditions, blocks=blocks)

    def exitFor_statement(self, ctx: ModelicaParser.For_statementContext):
        self.ast[ctx] = ast.ForStatement(
            indices=self.ast[ctx.for_indices()],
            statements=self.ast[ctx.statement_block()],
        )

    # EXPRESSIONS ===========================================================

    def exitSimple_expression(self, ctx: ModelicaParser.Simple_expressionContext):
        if len(ctx.expr()) > 1:
            if len(ctx.expr()) > 2:
                step = self.ast[ctx.expr()[2]]
            else:
                step = ast.Primary(value=1)
            self.ast[ctx] = ast.Slice(
                start=self.ast[ctx.expr()[0]], stop=self.ast[ctx.expr()[1]], step=step
            )
        else:
            self.ast[ctx] = self.ast[ctx.expr()[0]]

    def exitExpression_simple(self, ctx: ModelicaParser.Expression_simpleContext):
        self.ast[ctx] = self.ast[ctx.simple_expression()]

    def exitExpression_if(self, ctx: ModelicaParser.Expression_ifContext):
        all_expr = [self.ast[s] for s in ctx.expression()]
        # Note that an else block is guaranteed to exist.
        conditions = all_expr[:-1:2]
        expressions = all_expr[1::2] + all_expr[-1:]

        self.ast[ctx] = ast.IfExpression(conditions=conditions, expressions=expressions)

    def exitExpr_primary(self, ctx: ModelicaParser.Expr_primaryContext):
        self.ast[ctx] = self.ast[ctx.primary()]

    def exitExpr_add(self, ctx: ModelicaParser.Expr_addContext):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.expr()],
        )

    def exitExpr_exp(self, ctx: ModelicaParser.Expr_expContext):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.primary()],
        )

    def exitExpr_mul(self, ctx: ModelicaParser.Expr_mulContext):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.expr()],
        )

    def exitExpr_rel(self, ctx: ModelicaParser.Expr_relContext):
        self.ast[ctx] = ast.Expression(
            operator=ctx.op.text,
            operands=[self.ast[e] for e in ctx.expr()],
        )

    def exitExpr_not(self, ctx: ModelicaParser.Expr_notContext):
        self.ast[ctx] = ast.Expression(operator="not", operands=[self.ast[ctx.expr()]])

    def exitExpr_and(self, ctx: ModelicaParser.Expr_andContext):
        self.ast[ctx] = ast.Expression(
            operator="and",
            operands=[self.ast[e] for e in ctx.expr()],
        )

    def exitExpr_or(self, ctx: ModelicaParser.Expr_orContext):
        self.ast[ctx] = ast.Expression(
            operator="or",
            operands=[self.ast[e] for e in ctx.expr()],
        )

    def exitExpr_signed(self, ctx: ModelicaParser.Expr_signedContext):
        self.ast[ctx] = ast.Expression(operator=ctx.op.text, operands=[self.ast[ctx.expr()]])

    def exitSubscript(self, ctx: ModelicaParser.SubscriptContext):
        if ctx.expression() is not None:
            self.ast[ctx] = self.ast[ctx.expression()]
        else:
            self.ast[ctx] = ast.Slice()

    def exitArray_subscripts(self, ctx: ModelicaParser.Array_subscriptsContext):
        self.ast[ctx] = [self.ast[s] for s in ctx.subscript()]

    def exitFor_index(self, ctx):
        self.ast[ctx] = ast.ForIndex(
            name=ctx.IDENT().getText(), expression=self.ast[ctx.expression()]
        )

    def exitFor_indices(self, ctx: ModelicaParser.For_indicesContext):
        self.ast[ctx] = [self.ast[s] for s in ctx.for_index()]

    # PRIMARY ===========================================================

    def exitPrimary_unsigned_number(self, ctx: ModelicaParser.Primary_unsigned_numberContext):
        number_string = ctx.getText()
        try:
            val = int(number_string)
        except ValueError:
            val = float(number_string)

        self.ast[ctx] = ast.Primary(value=val)

    def exitPrimary_string(self, ctx: ModelicaParser.Primary_stringContext):
        val = ctx.getText()
        assert val.startswith('"') and val.endswith('"')
        self.ast[ctx] = ast.Primary(value=val[1:-1])

    def exitPrimary_false(self, ctx: ModelicaParser.Primary_falseContext):
        self.ast[ctx] = ast.Primary(value=False)

    def exitPrimary_true(self, ctx: ModelicaParser.Primary_trueContext):
        self.ast[ctx] = ast.Primary(value=True)

    def exitPrimary_function(self, ctx: ModelicaParser.Primary_functionContext):
        # TODO: Could possible be cleaner if we let the expression in the ast bubble up.
        #       E.g. self.ast[x] below, instead of self.ast[x.expression].
        self.ast[ctx] = ast.Expression(
            operator=self.ast[ctx.component_reference()],
            operands=[
                self.ast[x.expression()]
                for x in ctx.function_call_args().function_arguments().function_argument()
            ],
        )

    def exitPrimary_derivative(self, ctx: ModelicaParser.Primary_derivativeContext):
        self.ast[ctx] = ast.Expression(
            operator="der",
            operands=[
                self.ast[x.expression()]
                for x in ctx.function_call_args().function_arguments().function_argument()
            ],
        )
        # TODO 'state' is not a standard prefix;  disable this for now as it does not work
        # when differentiating states defined in superclasses.
        # if 'state' not in self.class_node.symbols[comp_name].prefixes:
        #    self.class_node.symbols[comp_name].prefixes += ['state']

    def exitType_specifier_element(self, ctx: ModelicaParser.Type_specifier_elementContext):
        self.ast[ctx] = ast.ComponentRef(name=ctx.IDENT().getText(), indices=[[None]], child=[])

    def exitType_specifier(self, ctx: ModelicaParser.Type_specifierContext):
        for element in reversed([self.ast[x] for x in ctx.type_specifier_element()]):
            if ctx in self.ast:
                element.child = [self.ast[ctx]]
            self.ast[ctx] = element

    def exitComponent_reference_element(
        self, ctx: ModelicaParser.Component_reference_elementContext
    ):
        if ctx.array_subscripts() is not None:
            indices = [[self.ast[x] for x in ctx.array_subscripts().subscript()]]
        else:
            indices = [[None]]
        self.ast[ctx] = ast.ComponentRef(name=ctx.IDENT().getText(), indices=indices, child=[])

    def exitComponent_reference(self, ctx: ModelicaParser.Component_referenceContext):
        for element in reversed([self.ast[x] for x in ctx.component_reference_element()]):
            if ctx in self.ast:
                element.child = [self.ast[ctx]]
            self.ast[ctx] = element

    def exitPrimary_component_reference(
        self, ctx: ModelicaParser.Primary_component_referenceContext
    ):
        self.ast[ctx] = self.ast[ctx.component_reference()]

    def exitPrimary_output_expression_list(
        self, ctx: ModelicaParser.Primary_output_expression_listContext
    ):
        self.ast[ctx] = [self.ast[x] for x in ctx.output_expression_list().expression()]
        # Collapse lists containing a single expression
        if len(self.ast[ctx]) == 1:
            self.ast[ctx] = self.ast[ctx][0]

    def exitPrimary_function_arguments(self, ctx: ModelicaParser.Primary_function_argumentsContext):
        # TODO: This does not support for generators yet.
        #       Only expressions are supported, e.g. {1.0, 2.0, 3.0}.
        v = [self.ast[x.expression()] for x in ctx.function_arguments().function_argument()]
        self.ast[ctx] = ast.Array(values=v)

    def exitEquation_function(self, ctx: ModelicaParser.Equation_functionContext):
        self.ast[ctx] = ast.Function(
            name=ctx.name().getText(),
            arguments=[
                self.ast[x.expression()]
                for x in ctx.function_call_args().function_arguments().function_argument()
            ],
        )

    def exitEquation_when(self, ctx: ModelicaParser.Equation_whenContext):
        self.ast[ctx] = self.ast[ctx.when_equation()]

    # COMPONENTS ===========================================================

    def exitElement_list(self, ctx: ModelicaParser.Element_listContext):
        self.ast[ctx] = [self.ast[e] for e in ctx.element()]

    def exitElement(self, ctx: ModelicaParser.ElementContext):
        self.ast[ctx] = self.ast[ctx.getChild(ctx.getAltNumber())]

    # TODO: Clean this up (inheritance or different import clause classes?)
    def exitImport_clause(self, ctx: ModelicaParser.Import_clauseContext):
        import_clause = ast.ImportClause()
        self.ast[ctx] = import_clause
        import_clause.components = [self.ast[ctx.component_reference()]]
        if ctx.IDENT() is not None:
            import_clause.short_name = ctx.IDENT().getText()
        else:
            import_list = ctx.import_list()
            if import_list is not None:
                package_name = import_clause.components.pop()
                # Append list of names to package_name to get fully qualified name(s)
                # Skip the comma separators in import_list.children
                for ident in import_list.children[::2]:
                    qualified_name = package_name.concatenate(
                        package_name.from_string(ident.getText())
                    )
                    import_clause.components.append(qualified_name)
            elif ctx.getChildCount() > 3:
                import_clause.unqualified = True
        if import_clause.short_name:
            # import_clause instead of comp_ref signifies short_name
            self.class_node.imports[import_clause.short_name] = import_clause
        elif import_clause.unqualified:
            # Postpone processing this uncommon case until actually needed
            # In this case import_clause.components contains list of packages of all unqualified imports
            if "*" not in self.class_node.imports:
                self.class_node.imports["*"] = import_clause
            else:
                self.class_node.imports["*"].components.append(import_clause.components[0])
        else:
            # Simple case, fast lookup
            for comp in import_clause.components:
                name = comp.to_tuple()[-1]
                # Check for name clashes
                if name in self.class_node.imports:
                    raise IOError(name, "already imported")
                self.class_node.imports[name] = comp

    def enterExtends_clause(self, ctx: ModelicaParser.Extends_clauseContext):
        self.in_extends_clause = True

    def exitExtends_clause(self, ctx: ModelicaParser.Extends_clauseContext):
        if ctx.class_modification() is not None:
            class_modification = self.ast[ctx.class_modification()]
        else:
            class_modification = ast.ClassModification()
        self.ast[ctx] = ast.ExtendsClause(
            component=self.ast[ctx.component_reference()], class_modification=class_modification
        )
        self.class_node.extends += [self.ast[ctx]]

        self.in_extends_clause = False

    def exitRegular_element(self, ctx: ModelicaParser.Regular_elementContext):
        if ctx.comp_elem is not None:
            self.ast[ctx] = self.ast[ctx.comp_elem]
        else:
            self.ast[ctx] = self.ast[ctx.class_elem]

    def exitReplaceable_element(self, ctx: ModelicaParser.Replaceable_elementContext):
        if ctx.comp_elem is not None:
            self.ast[ctx] = self.ast[ctx.comp_elem]
        else:
            self.ast[ctx] = self.ast[ctx.class_elem]

    def enterComponent_clause(self, ctx: ModelicaParser.Component_clauseContext):
        prefixes = ctx.type_prefix().getText().split(" ")
        if prefixes[0] == "":
            prefixes = []
        self.ast[ctx] = ast.ComponentClause(
            prefixes=prefixes,
        )
        self.comp_clause = self.ast[ctx]

    def enterComponent_clause1(self, ctx: ModelicaParser.Component_clause1Context):
        prefixes = ctx.type_prefix().getText().split(" ")
        if prefixes[0] == "":
            prefixes = []
        self.ast[ctx] = ast.ComponentClause(
            prefixes=prefixes,
        )
        self.comp_clause = self.ast[ctx]

    def exitComponent_clause(self, ctx: ModelicaParser.Component_clauseContext):
        clause = self.ast[ctx]
        # The component clause and all its symbols share the same type.
        # However, the type will only be turned into a component reference
        # somewhere between the enterDeclaration and exitDeclaration functions
        # of the symbols. Therefore, we need to keep the component clause's
        # type, and all its symbols' types, pointing at the same empty
        # (ComponentRef) object until we can fill it.
        clause.type.__dict__.update(self.ast[ctx.type_specifier()].__dict__)
        if ctx.array_subscripts() is not None:
            clause.dimensions = [self.ast[ctx.array_subscripts()]]
            for sym in self.comp_clause.symbol_list:
                s = self.class_node.symbols[sym.name]
                s.dimensions = clause.dimensions

        # We make sure that all references to the objects are unique per
        # symbol making copies. Note that if there is only one symbol in the
        # component clause, it is already unique.
        for sym in self.comp_clause.symbol_list[1:]:
            s = self.class_node.symbols[sym.name]
            s.dimensions = list(s.dimensions)
            s.prefixes = list(s.prefixes)
            s.type = copy.deepcopy(clause.type)

    def exitComponent_clause1(self, ctx: ModelicaParser.Component_clause1Context):
        clause = self.ast[ctx]
        clause.type.__dict__.update(self.ast[ctx.type_specifier()].__dict__)

        for sym in self.comp_clause.symbol_list[1:]:
            s = self.class_node.symbols[sym.name]
            s.dimensions = list(s.dimensions)
            s.prefixes = list(s.prefixes)
            s.type = copy.deepcopy(clause.type)

    def enterComponent_declaration(self, ctx: ModelicaParser.Component_declarationContext):
        sym = ast.Symbol(order=self.sym_count)
        self.sym_count += 1
        self.ast[ctx] = sym
        self.symbol_node = sym
        self.comp_clause.symbol_list += [sym]

    def enterComponent_declaration1(self, ctx: ModelicaParser.Component_declaration1Context):
        sym = ast.Symbol(order=self.sym_count)
        self.sym_count += 1
        self.ast[ctx] = sym
        self.symbol_node = sym
        self.comp_clause.symbol_list += [sym]

    def enterElement_modification(self, ctx: ModelicaParser.Element_modificationContext):
        if self.symbol_node is not None:
            self.ast[ctx] = self.symbol_node
        else:
            sym = ast.Symbol(order=self.sym_count)
            self.sym_count += 1
            self.ast[ctx] = sym
            self.symbol_node = sym

    def exitComponent_declaration(self, ctx: ModelicaParser.Component_declarationContext):
        self.ast[ctx].comment = self.ast[ctx.comment()]
        self.symbol_node = None

    def exitComponent_declaration1(self, ctx: ModelicaParser.Component_declaration1Context):
        self.ast[ctx].comment = self.ast[ctx.comment()]
        self.symbol_node = None

    def enterDeclaration(self, ctx: ModelicaParser.DeclarationContext):
        sym = self.symbol_node
        dimensions = None
        if self.comp_clause.dimensions is not None:
            dimensions = self.comp_clause.dimensions
        sym.name = ctx.IDENT().getText()
        sym.dimensions = dimensions
        sym.prefixes = self.comp_clause.prefixes
        sym.type = self.comp_clause.type

        # Declarations can also occur in extends clauses, in which case we do not have to add it to the class's symbols.
        if not self.in_extends_clause:
            if sym.name in self.class_node.symbols:
                raise IOError(sym.name, "already defined")
            self.class_node.symbols[sym.name] = sym

    def exitDeclaration(self, ctx: ModelicaParser.DeclarationContext):
        sym = self.symbol_node
        if ctx.array_subscripts() is not None:
            sym.dimensions = [self.ast[ctx.array_subscripts()]]
        if ctx.modification() is not None:
            for mod in self.ast[ctx.modification()]:
                if isinstance(mod, ast.ClassModification):
                    sym.class_modification = mod
                else:
                    # Assignment of value, which we turn into a modification here.
                    vmod_arg = ast.ClassModificationArgument()
                    vmod_arg.value = ast.ElementModification()
                    vmod_arg.value.component = ast.ComponentRef(name="value")
                    vmod_arg.value.modifications = [mod]

                    if sym.class_modification is None:
                        sym_mod = ast.ClassModification()
                        sym_mod.arguments.append(vmod_arg)
                        sym.class_modification = sym_mod
                    else:
                        sym.class_modification.arguments.append(vmod_arg)

    def exitElement_modification(self, ctx: ModelicaParser.Element_modificationContext):
        component = self.ast[ctx.component_reference()]
        if ctx.modification() is not None:
            modifications = self.ast[ctx.modification()]
        else:
            modifications = []

        self.ast[ctx] = ast.ElementModification(component=component, modifications=modifications)

    def exitModification_class(self, ctx: ModelicaParser.Modification_classContext):
        self.ast[ctx] = [self.ast[ctx.class_modification()]]
        if ctx.expression() is not None:
            self.ast[ctx] += [self.ast[ctx.expression()]]

    def exitModification_assignment(self, ctx: ModelicaParser.Modification_assignmentContext):
        self.ast[ctx] = [self.ast[ctx.expression()]]

    def exitModification_assignment2(self, ctx: ModelicaParser.Modification_assignment2Context):
        self.ast[ctx] = [self.ast[ctx.expression()]]

    def exitElement_replaceable(self, ctx: ModelicaParser.Element_replaceableContext):
        if ctx.component_clause1() is not None:
            self.ast[ctx] = self.ast[ctx.component_clause1()]
        else:
            self.ast[ctx] = self.ast[ctx.short_class_definition()]

    def exitElement_modification_or_replaceable(
        self, ctx: ModelicaParser.Element_modification_or_replaceableContext
    ):
        if ctx.element_modification() is not None:
            self.ast[ctx] = self.ast[ctx.element_modification()]
        else:
            self.ast[ctx] = self.ast[ctx.element_replaceable()]

    def exitElement_redeclaration(self, ctx: ModelicaParser.Element_redeclarationContext):
        if ctx.component_clause1() is not None:
            self.ast[ctx] = self.ast[ctx.component_clause1()]
        else:
            self.ast[ctx] = self.ast[ctx.short_class_definition()]

    # COMMENTS ==============================================================

    def exitComment(self, ctx: ModelicaParser.CommentContext):
        self.ast[ctx] = self.ast[ctx.string_comment()]

    def exitString_comment(self, ctx: ModelicaParser.String_commentContext):
        self.ast[ctx] = ctx.getText()[1:-1]

    # ANNOTATIONS ==========================================================

    def exitAnnotation(self, ctx: ModelicaParser.AnnotationContext):
        self.ast[ctx] = self.ast[ctx.class_modification()]


# UTILITY FUNCTIONS ========================================================
def file_to_tree(f: ModelicaFile) -> ast.Tree:
    # TODO: We can only insert where classes exist. For example, if we have a
    # within statement, we have to check if the nodes of the within statement
    # are actually in the tree, and if not raise an exception.
    root = ast.Tree()
    insert_node = root
    if f.within:
        for p in f.within[0].to_tuple():
            package = ast.Class(name=p, type="package")
            insert_node.classes[p] = package
            insert_node = package

    insert_node.classes.update(f.classes)

    root.update_parent_refs()

    return root


class ModelicaParserErrorListener(antlr4.error.ErrorListener.ErrorListener):
    def __init__(self):
        self._error = False
        super().__init__()

    @property
    def error(self):
        return self._error

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        self._error = True
        super().syntaxError(recognizer, offendingSymbol, line, column, msg, e)


def _parse(text: str) -> Union[ast.Tree, None]:
    """Parse Modelica code given in text"""
    input_stream = antlr4.InputStream(text)
    lexer = ModelicaLexer(input_stream)
    stream = antlr4.CommonTokenStream(lexer)
    parser = ModelicaParser(stream)
    # parser.buildParseTrees = False
    listener = ModelicaParserErrorListener()
    parser.addErrorListener(listener)
    parse_tree = parser.stored_definition()
    if listener.error:
        return None
    ast_listener = ASTListener()
    parse_walker = antlr4.ParseTreeWalker()
    parse_walker.walk(ast_listener, parse_tree)
    modelica_file = ast_listener.file_node
    return file_to_tree(modelica_file)


def _microseconds_since_epoch(timedelta_: Optional[timedelta] = None) -> int:
    if timedelta_ is None:
        timedelta_ = timedelta()
    return time.time_ns() // 1000 + int(timedelta_.total_seconds() * 1e6)


def _check_database_structure(conn: sqlite3.Connection):
    """
    Function to check if the existing database file matches the expected table structure
    """
    cursor = conn.cursor()

    cursor.execute("BEGIN TRANSACTION;")
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='models'")
    table_exists = cursor.fetchone()
    table_correct = False

    if table_exists:
        cursor.execute("PRAGMA table_info('models')")
        columns = cursor.fetchall()
        expected_columns = [
            (0, "txt_hash", "TEXT", 0, None, 1),
            (1, "pymoca_version", "TEXT", 0, None, 2),
            (2, "data", "BLOB", 0, None, 0),
            (3, "last_hit", "TIMESTAMP INTEGER", 0, None, 0),
        ]

        if columns != expected_columns:
            logger.warning("Model text cache table layout didn't match, recreating")
            table_correct = False
        else:
            table_correct = True

    if not table_correct:
        cursor.execute("DROP TABLE IF EXISTS models")

        logger.debug("Creating model text cache table in database")
        cursor.execute(
            """
            CREATE TABLE models (
                txt_hash TEXT,
                pymoca_version TEXT,
                data BLOB,
                last_hit TIMESTAMP INTEGER,
                PRIMARY KEY (txt_hash, pymoca_version)
            )
        """
        )

    conn.commit()

    # For metadata we check if the table layout is correct, but also whether
    # the metadata keys exist.
    cursor.execute("BEGIN TRANSACTION;")
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table' AND name='metadata'")
    metadata_table_exists = cursor.fetchone()
    metadata_table_correct = False

    if metadata_table_exists:
        cursor.execute("PRAGMA table_info('metadata')")
        columns = cursor.fetchall()
        expected_columns = [
            (0, "key", "TEXT", 0, None, 1),
            (1, "value", "TEXT", 0, None, 0),
        ]

        if columns != expected_columns:
            logger.warning("Metadata table layout didn't match, recreating")
            metadata_table_correct = False
        else:
            metadata_table_correct = True

    if not metadata_table_correct:
        cursor.execute("DROP TABLE IF EXISTS metadata")

        logger.debug("Creating metadata table in database")
        cursor.execute(
            """
            CREATE TABLE metadata (
                key TEXT,
                value TEXT,
                PRIMARY KEY (key)
            )
        """
        )
    conn.commit()

    cursor.execute("BEGIN TRANSACTION;")
    cursor.execute(
        "INSERT OR IGNORE INTO metadata (key, value) VALUES (?, ?)",
        ("created_at", _microseconds_since_epoch()),
    )

    cursor.execute(
        "INSERT OR IGNORE INTO metadata (key, value) VALUES (?, ?)",
        ("last_prune", _microseconds_since_epoch()),
    )

    conn.commit()


def _calculate_txt_hash(txt: str):
    hasher = hashlib.sha256()
    hasher.update(txt.encode("utf-8"))
    return hasher.hexdigest()


def _get_default_cache_path() -> Path:
    """
    Heavily inspired by https://github.com/davidhalter/parso/blob/master/parso/cache.py

    The path where the cache is stored.
    On Linux, this defaults to ``~/.cache/pymoca/``, on OS X to
    ``~/Library/Caches/Pymoca/`` and on Windows to ``%LOCALAPPDATA%\\Pymoca\\Pymoca\\``.
    On Linux, if environment variable ``$XDG_CACHE_HOME`` is set,
    ``$XDG_CACHE_HOME/pymoca`` is used instead of the default one.
    """
    if platform.system().lower() == "windows":
        dir_ = Path(os.getenv("LOCALAPPDATA") or "~") / "Pymoca" / "Pymoca"
    elif platform.system().lower() == "darwin":
        dir_ = Path("~") / "Library" / "Caches" / "Pymoca"
    else:
        dir_ = Path(os.getenv("XDG_CACHE_HOME") or "~/.cache") / "pymoca"

    return dir_.expanduser()


def parse(
    txt: str,
    /,
    model_cache_folder: Optional[Path] = None,
    cache_db: str = DEFAULT_MODEL_CACHE_DB,
    cache_expiration_days: int = 30,
    always_update_last_hit: bool = False,
    bypass_cache: bool = False,
) -> Union[ast.Tree, None]:
    """
    Parse the Modelica code given in text and return the Abstract Syntax Tree (AST).

    This function uses a cache to avoid re-parsing the same text multiple times. The
    cache is stored in a SQLite database file. The cache entries are pruned based on
    their last access time. If an entry has not been accessed for a certain number of
    days (defined by cache_expiration_days), it is removed from the cache.

    Args:
        txt (str): The Modelica code to parse.
        model_cache_folder (Path, optional): The folder where the cache database is
            stored. If not provided, a default location based on the operating system is
            used.
        cache_db (str, optional): The name of the cache database file. Default is
            DEFAULT_MODEL_CACHE_DB.
        cache_expiration_days (int, optional): The number of days after which a cache
            entry is considered expired and is pruned. Defaultis 30.
        always_update_last_hit (bool, optional): If True, the last access time of a
            cache entry is always updated when it is accessed. If False, the last access
            time is only updated if it is more than a day old. Default is False.
        bypass_cache (bool, optional): If True, the cache is bypassed and the parsing is
            performed directly. Default is False.

    Returns:
        Union[ast.Tree, None]: The AST of the parsed Modelica code, or None if the
            parsing failed.
    """
    if bypass_cache:
        return _parse(txt)

    pymoca_version = pymoca.__version__

    # Do not use caching if we have a dirty work tree, as the source
    # code can't be uniquely identified.
    if pymoca_version.endswith(".dirty"):
        logger.debug("Bypassing cache because working directory is dirty")
        return _parse(txt)

    if model_cache_folder is not None:
        db_folder = model_cache_folder
    else:
        db_folder = _get_default_cache_path()
    db_folder.mkdir(parents=True, exist_ok=True)

    full_db_path = db_folder / cache_db
    conn = sqlite3.connect(full_db_path, isolation_level=None)

    cursor = conn.cursor()

    if not hasattr(parse, "initialized_dbs") or full_db_path not in parse.initialized_dbs:
        # Check if the database file is corrupt
        try:
            cursor.execute("PRAGMA integrity_check;")
            result = cursor.fetchone()
            if result != ("ok",):
                raise sqlite3.DatabaseError("Database integrity check failed")
        except sqlite3.DatabaseError:
            conn.close()

            logger.warning("Model cache database is corrupt, recreating...")
            os.remove(full_db_path)

            conn = sqlite3.connect(full_db_path, isolation_level=None)
            cursor = conn.cursor()

        _check_database_structure(conn)

        # Prune the database of entries not hit recently
        cursor.execute("BEGIN TRANSACTION;")
        cutoff_time = _microseconds_since_epoch(timedelta(days=-cache_expiration_days))
        cursor.execute("DELETE FROM models WHERE last_hit < ?", (cutoff_time,))
        # Sometimes Windows time resolution is a bit coarse, so we make
        # sure that if we update the last_prune time, it is actually newer
        # than the previous one.
        cursor.execute(
            "UPDATE metadata SET value = max(value + 1, ?) WHERE key = ?",
            (_microseconds_since_epoch(), "last_prune"),
        )

        conn.commit()

        if hasattr(parse, "initialized_dbs"):
            parse.initialized_dbs.add(full_db_path)
        else:
            parse.initialized_dbs = {full_db_path}

    # Check if the txt exists in the database
    txt_hash = _calculate_txt_hash(txt)

    cursor.execute("BEGIN TRANSACTION;")
    cursor.execute(
        "SELECT last_hit, data FROM models WHERE txt_hash=? AND pymoca_version=?",
        (txt_hash, pymoca_version),
    )
    result = cursor.fetchone()
    conn.commit()

    tree = None

    if result:
        logger.debug(f"Model with hash '{txt_hash}' ({pymoca_version}) found in cache")
        last_hit, pickled_data = result

        yesterday = _microseconds_since_epoch(timedelta(days=-1))

        if always_update_last_hit or last_hit < yesterday:
            cursor.execute("BEGIN TRANSACTION;")
            # Sometimes Windows time resolution is a bit coarse, so we make
            # sure that if we update the last_hit time, it is actually newer
            # than the previous one.
            cursor.execute(
                "UPDATE models SET last_hit = max(last_hit + 1, ?) WHERE txt_hash = ? "
                "AND pymoca_version = ?",
                (_microseconds_since_epoch(), txt_hash, pymoca_version),
            )
            conn.commit()
        try:
            tree = pickle.loads(pickled_data)
        except pickle.UnpicklingError:
            logger.warning(f"Model with hash '{txt_hash}' ({pymoca_version}) failed to unpickle")
    else:
        logger.debug(f"Model with hash '{txt_hash}' ({pymoca_version}) not in cache")

    if tree is None:
        # We get here if we didn't find anything in the cache, or if the
        # unpickling of the cache failed
        try:
            tree = _parse(txt)
        except Exception:
            conn.close()
            raise

        # Don't cache None that _parse() returns on syntax errors
        if tree is not None:
            pickled_data = pickle.dumps(tree)

            # Note that we do an 'INSERT OR REPLACE' because concurrent access
            # might mean two processes/threads try to insert an entry
            cursor.execute("BEGIN TRANSACTION;")
            cursor.execute(
                "INSERT OR REPLACE INTO models (txt_hash, pymoca_version, data, last_hit) VALUES (?, ?, ?, ?)",
                (txt_hash, pymoca_version, pickled_data, _microseconds_since_epoch()),
            )
            conn.commit()

    conn.close()

    return tree



import os
from collections import OrderedDict
from typing import Union

import casadi as ca

from lxml import etree

from .model import HybridDae

FILE_PATH = os.path.dirname(os.path.abspath(__file__))
SCHEMA_DIR = os.path.join(FILE_PATH, "ModelicaXML", "schemas")


class XMLParser:
    def __init__(self, schema_dir, schema_file):
        orig_path = os.path.abspath(os.curdir)
        os.chdir(schema_dir)
        with open(schema_file, "r") as f:
            schema = etree.XMLSchema(etree.XML(f.read().encode("utf-8")))
        os.chdir(orig_path)
        self._parser = etree.XMLParser(schema=schema, remove_comments=True, remove_blank_text=False)

    def parse(self, txt: str):
        if not isinstance(txt, str):
            raise ValueError("txt must be a str")
        xml_file = txt.encode("utf-8")
        return etree.fromstring(xml_file, self._parser)


Sym = Union[ca.MX, ca.SX]


class ModelListener:
    """Converts ModelicaXML file to Hybrid DAE"""

    def __init__(self, sym: Sym = ca.SX, verbose=False):
        self.depth = 0
        self.model = {}
        self.scope_stack = []
        self.verbose = verbose
        self.sym = sym

        # Define an operator map that can be used as
        # self.op_map[n_operations][operator](*args)
        self.op_map = {
            1: {
                "der": self.der,
                "-": lambda x: -1 * x,
                "abs": ca.fabs,
                "sin": ca.sin,
                "cos": ca.cos,
                "tan": ca.tan,
            },
            2: {
                "+": lambda x, y: x + y,
                "-": lambda x, y: x - y,
                "*": lambda x, y: x * y,
                "/": lambda x, y: x / y,
                "^": lambda x, y: ca.power(x, y),
                ">": lambda x, y: x > y,
                "<": lambda x, y: x < y,
                "<=": lambda x, y: x <= y,
                ">=": lambda x, y: x >= y,
                "reinit": self.reinit,
                "sample": self.sample,
                "and": ca.logic_and,
                "or": ca.logic_or,
                "not": ca.logic_not,
                "noise_gaussian": lambda mean, std: self.noise_gaussian(mean, std),
                "noise_uniform": lambda lower, upper: self.noise_uniform(lower, upper),
            },
        }

    @property
    def scope(self):
        return self.scope_stack[-1]

    def call(self, tag_name: str, *args, **kwargs):
        """Convenience method for calling methods with walker."""
        if hasattr(self, tag_name):
            getattr(self, tag_name)(*args, **kwargs)

    # ------------------------------------------------------------------------
    # OPERATORS
    # ------------------------------------------------------------------------

    def der(self, x: Sym):
        """Get the derivative of the variable, create it if it doesn't exist."""
        name = "der({:s})".format(x.name())
        if name not in self.scope["dvar"].keys():
            self.scope["dvar"][name] = self.sym.sym(name, *x.shape)
            self.scope["states"].append(x.name())
        return self.scope["dvar"][name]

    def cond(self, expr):
        c = self.sym.sym("c_{:d}".format(len(self.scope["c"])))
        self.scope["c"][c] = expr
        return c

    def pre_cond(self, x: Sym):
        name = "pre({:s})".format(x.name())
        if name not in self.scope["pre_c"].keys():
            self.scope["pre_c"][name] = self.sym.sym(name, *x.shape)
        return self.scope["pre_c"][name]

    def edge(self, c):
        """rising edge"""
        return ca.logic_and(c, ca.logic_not(self.pre_cond(c)))

    @staticmethod
    def reinit(x_old, x_new):
        return "reinit", x_old, x_new

    @staticmethod
    def sample(t_start, period):
        print("sample", t_start, period)
        return "sample", t_start, period

    def noise_gaussian(self, mean, std):
        """Create a gaussian noise variable"""
        assert std > 0
        ng = self.sym.sym("ng_{:d}".format(len(self.scope["ng"])))
        self.scope["ng"].append(ng)
        return mean + std * ng

    def noise_uniform(self, lower_bound, upper_bound):
        """Create a uniform noise variable"""
        assert upper_bound > lower_bound
        nu = self.sym.sym("nu_{:d}".format(len(self.scope["nu"])))
        self.scope["nu"].append(nu)
        return lower_bound + nu * (upper_bound - lower_bound)

    # ------------------------------------------------------------------------
    # Helpers
    # ------------------------------------------------------------------------

    @staticmethod
    def get_attr(e, name, default):
        if name in e.attrib.keys():
            return e.attrib[name]
        else:
            return default

    def get_var(self, name):
        """Get the variable in the current scope"""
        if name == "time":
            return self.scope["time"]
        else:
            return self.scope["var"][name]

    def log(self, *args, **kwargs):
        """Convenience function for printing indenting debug output."""
        if self.verbose:
            print("   " * self.depth, *args, **kwargs)

    # ------------------------------------------------------------------------
    # Listener Methods
    # ------------------------------------------------------------------------

    def enter_every_before(self, tree: etree._Element):
        # initialize model to None
        self.model[tree] = None

        # print name to log
        self.log(tree.tag, "{")

        # increment depth
        self.depth += 1

    def exit_every_after(self, tree: etree._Element):
        # decrement depth
        self.depth -= 1

        # self.log('tree:', etree.tostring(tree))

        # print model
        if self.model[tree] is not None:
            self.log("model:", self.model[tree])

        # print name to log
        self.log("}", tree.tag)

    def enter_classDefinition(self, tree: etree._Element):
        # we don't know if variables are states
        # yet, we need to wait until equations are parsed
        self.scope_stack.append(
            {
                "time": self.sym.sym("time"),
                "sample_times": [],
                "var": OrderedDict(),  # variables
                "states": [],  # list of which variables are states (based on der call)
                "dvar": OrderedDict(),  # derivative of variables
                "eqs": [],  # equations
                "when_eqs": [],  # when equations
                "c": {},  # conditions
                "pre_c": {},  # pre conditions
                "p": [],  # parameters and constants
                "prop": {},  # properties for variables
                "ng": [],  # gaussian
                "nu": [],  # uniform
            }
        )

    def exit_classDefinition(self, tree: etree._Element):
        dae = HybridDae()
        dae.t = self.scope["time"]
        self.model[tree] = dae

        # handle component declarations
        for var_name, v in self.scope["var"].items():
            variability = self.scope["prop"][var_name]["variability"]
            if variability == "continuous":
                if var_name in self.scope["states"]:
                    dae.x = ca.vertcat(dae.x, v)
                    dae.dx = ca.vertcat(dae.dx, self.der(v))
                else:
                    dae.y = ca.vertcat(dae.y, v)
            elif variability == "discrete":
                dae.m = ca.vertcat(dae.m, v)
            elif variability == "parameter":
                dae.p = ca.vertcat(dae.p, v)
            elif variability == "constant":
                dae.p = ca.vertcat(dae.p, v)
            else:
                raise ValueError("unknown variability", variability)

        for eq in self.scope["eqs"]:
            if isinstance(eq, self.sym):
                dae.f_x = ca.vertcat(dae.f_x, eq)

        # build reinit expression and discrete equations
        dae.f_i = dae.x
        dae.f_m = dae.m
        for eq in self.scope["when_eqs"]:
            w = eq["cond"]
            for then_eq in eq["then"]:
                if isinstance(then_eq, tuple):
                    if then_eq[0] == "reinit":
                        sub_var = then_eq[1]
                        sub_expr = ca.if_else(self.edge(w), then_eq[2], sub_var)
                        dae.f_i = ca.substitute(dae.f_i, sub_var, sub_expr)
                elif isinstance(then_eq, self.sym):
                    # this is a discrete variable assignment
                    # so it should be a casadi subtraction y = x
                    assert then_eq.is_op(ca.OP_SUB) and then_eq.n_dep() == 2
                    sub_var = then_eq.dep(0)
                    sub_expr = ca.if_else(self.edge(w), then_eq.dep(1), sub_var)
                    dae.f_m = ca.substitute(dae.f_m, sub_var, sub_expr)

        dae.t = self.scope["time"]
        dae.prop.update(self.scope["prop"])
        c_dict = self.scope["c"]
        for k in c_dict.keys():
            dae.c = ca.vertcat(dae.c, k)
            dae.pre_c = ca.vertcat(dae.pre_c, self.pre_cond(k))
            dae.f_c = ca.vertcat(dae.f_c, c_dict[k])

        for left, right in [("f_c", "c"), ("c", "pre_c"), ("dx", "x"), ("f_m", "m")]:
            vl = getattr(dae, left)
            vr = getattr(dae, right)
            if vl.shape != vr.shape:
                raise ValueError(
                    "{:s} and {:s} must have the same shape:"
                    "\n{:s}: {:s}\t{:s}: {:s}".format(
                        left, right, left, str(dae.f_m), right, str(dae.m)
                    )
                )

        dae.ng = ca.vertcat(*self.scope["ng"])
        dae.nu = ca.vertcat(*self.scope["nu"])

        n_eq = dae.f_x.shape[0] + dae.f_m.shape[0]
        n_var = dae.x.shape[0] + dae.m.shape[0] + dae.y.shape[0]
        if n_eq != n_var:
            raise ValueError(
                "must have equal number of equations "
                "{:d} and unknowns {:d}\n:{:s}".format(n_eq, n_var, str(dae))
            )
        self.scope_stack.pop()

    def enter_component(self, tree: etree._Element):
        self.model[tree] = {
            "start": None,
            "fixed": None,
            "value": None,
            "variability": self.get_attr(tree, "variability", "continuous"),
            "visibility": self.get_attr(tree, "visibility", "public"),
        }
        self.scope_stack.append(self.model[tree])

    def exit_component(self, tree: etree._Element):
        var_scope = self.scope_stack.pop()
        name = tree.attrib["name"]
        shape = (1, 1)
        sym = self.sym.sym(name, *shape)
        self.scope["prop"][name] = var_scope
        self.scope["var"][name] = sym

    def exit_local(self, tree: etree._Element):
        name = tree.attrib["name"]
        self.model[tree] = self.get_var(name)

    def exit_operator(self, tree: etree._Element):
        op = tree.attrib["name"]
        self.model[tree] = self.op_map[len(tree)][op](*[self.model[e] for e in tree])

    def exit_if(self, tree: etree._Element):
        assert len(tree) == 3
        cond = self.model[tree[0]]
        then_eq = self.model[tree[1]]
        else_eq = self.model[tree[2]]
        c = self.cond(cond)
        if len(then_eq) != len(else_eq):
            raise SyntaxError("then and else equations must have same number of statements")
        self.model[tree] = ca.if_else(c, then_eq[0], else_eq[0])

    def exit_apply(self, tree: etree._Element):
        op = tree.attrib["builtin"]
        self.model[tree] = self.op_map[len(tree)][op](*[self.model[e] for e in tree])

    def exit_equal(self, tree: etree._Element):
        assert len(tree) == 2
        self.model[tree] = self.model[tree[0]] - self.model[tree[1]]

    def exit_equation(self, tree: etree._Element):
        self.model[tree] = [self.model[c] for c in tree]
        self.scope["eqs"].extend(self.model[tree])

    def exit_modifier(self, tree: etree._Element):
        props = {}
        for e in tree:
            props.update(self.model[e])
        self.model[tree] = props
        self.scope.update(props)

    def exit_item(self, tree: etree._Element):
        assert len(tree) == 1
        self.model[tree] = {tree.attrib["name"]: self.model[tree[0]]}

    def exit_real(self, tree: etree._Element):
        self.model[tree] = float(tree.attrib["value"])

    def exit_true(self, tree: etree._Element):
        self.model[tree] = True

    def exit_false(self, tree: etree._Element):
        self.model[tree] = False

    def exit_modelica(self, tree: etree._Element):
        # get all class definitions as a list
        self.model[tree] = [self.model[c] for c in tree[0]]

    def exit_when(self, tree: etree._Element):
        assert len(tree) == 2
        cond = self.model[tree[0]]
        then = self.model[tree[1]]
        self.model[tree] = {"cond": self.cond(cond), "then": then}
        self.scope["when_eqs"].append(self.model[tree])

    def exit_cond(self, tree: etree._Element):
        assert len(tree) == 1
        self.model[tree] = self.model[tree[0]]

    def exit_then(self, tree: etree._Element):
        self.model[tree] = [self.model[c] for c in tree]

    def exit_else(self, tree: etree._Element):
        self.model[tree] = [self.model[c] for c in tree]


def walk(e: etree._Element, listener: ModelListener) -> None:
    tag = e.tag
    listener.call("enter_every_before", e)
    listener.call("enter_" + tag, e)
    listener.call("enter_every_after", e)
    for c in e.getchildren():
        walk(c, listener)
    listener.call("exit_every_before", e)
    listener.call("exit_" + tag, e)
    listener.call("exit_every_after", e)


def parse(model_txt: str, verbose: bool = False) -> HybridDae:
    parser = XMLParser(SCHEMA_DIR, "Modelica.xsd")
    root = parser.parse(model_txt)
    listener = ModelListener(verbose=verbose)
    walk(root, listener)
    return listener.model[root][0]


def parse_file(file_path: str, verbose: bool = False) -> HybridDae:
    with open(file_path, "r") as f:
        txt = f.read()
    return parse(txt, verbose)



def _get_default_options():
    """
    Returns a dictionary with the available compiler options and the default values

    Returns:
        default_options (Dict)

    """
    return {
        "library_folders": [],
        "verbose": False,
        "check_balanced": True,
        "mtime_check": True,
        "cache": False,
        "codegen": False,
        "expand_mx": False,
        "unroll_loops": True,
        "inline_functions": True,
        "expand_vectors": False,
        "resolve_parameter_values": False,
        "replace_parameter_expressions": False,
        "replace_constant_expressions": False,
        "eliminate_constant_assignments": False,
        "replace_parameter_values": False,
        "replace_constant_values": False,
        "eliminable_variable_expression": None,
        "factor_and_simplify_equations": False,
        "detect_aliases": False,
        "allow_derivative_aliases": True,
        "reduce_affine_expression": False,
    }


def _merge_default_options(options):
    if options is None:
        return _get_default_options()
    elif isinstance(options, dict):
        default_options = _get_default_options()
        default_options.update(options)
        return default_options
    else:
        raise TypeError("options must be of type dict")



"""
The functions and classes defined here are only supposed to be used internally
within Pymoca to work around the <= 2D only shortcoming of CasADi. In the
eventual generated model, everything should should be ca.MX.
"""

import casadi as ca

import numpy as np


def _new_mx(name, *shape):
    if len(shape) == 1 and not np.isscalar(shape[0]):
        shape = shape[0]

    assert len(shape) <= 2

    obj = ca.MX.sym(name, *shape)
    obj._modelica_shape = tuple(shape)
    return obj


class _MTensor:
    def __init__(self, name, *shape):
        self._shape = tuple(shape)
        self._mx = ca.MX.sym(name, np.prod(shape))

    @property
    def shape(self):
        return self._shape

    def __getattr__(self, attr):
        return getattr(self._mx, attr)

    def __getitem__(self, k):
        assert len(k) == len(self._shape)
        return self._mx[
            np.ravel_multi_index(
                tuple(
                    k,
                ),
                self._shape,
            )
        ]



import itertools
import logging
import re
import sys
from collections import OrderedDict, namedtuple

import casadi as ca

import numpy as np

from ._options import _merge_default_options
from .alias_relation import AliasRelation
from .mtensor import _MTensor

logger = logging.getLogger("pymoca")

CASADI_COMPARISON_DEPTH = sys.maxsize
SUBSTITUTE_LOOP_LIMIT = 100
SIMPLIFICATION_LOOP_LIMIT = 50
CASADI_ATTRIBUTES = ("value", "min", "max", "start", "fixed", "nominal")


class _DefaultValue(int):
    pass


class Variable:
    def __init__(self, symbol, python_type=float, aliases=None):
        if aliases is None:
            aliases = set()
        self.symbol = symbol
        self.python_type = python_type
        self.aliases = aliases

        # Default attribute values
        self.value = np.nan
        self.start = _DefaultValue(0)
        self.min = -np.inf
        self.max = np.inf
        self.nominal = 0
        self.fixed = False

    def __str__(self):
        return self.symbol.name()

    def __repr__(self):
        return "{}[{},{}]:{}".format(
            self.symbol.name(), self.symbol.size1(), self.symbol.size2(), self.python_type.__name__
        )

    def to_dict(self):
        d = {}
        d["name"] = self.symbol.name()
        d["shape"] = (self.symbol.size1(), self.symbol.size2())
        d["python_type"] = self.python_type
        d["aliases"] = self.aliases

        for attr in CASADI_ATTRIBUTES:
            d[attr] = getattr(self, attr)
            if isinstance(d[attr], ca.MX):
                d[attr] = None  # Will be filled using variable_metadata_function

        return d

    @classmethod
    def from_dict(cls, d):
        variable = cls(ca.MX.sym(d["name"], *d["shape"]), d["python_type"])
        variable.aliases = d["aliases"]

        for attr in CASADI_ATTRIBUTES:
            setattr(variable, attr, d[attr])

        return variable


class StringVariable:
    def __init__(self, name):
        self.name = name
        self.value = None
        self.start = ""
        self.fixed = False

    def __str__(self):
        return self.name

    def __repr__(self):
        return "{}:str".format(self.name)

    def to_dict(self):
        return {k: getattr(self, k) for k in ["name", "value", "start", "fixed"]}

    @classmethod
    def from_dict(cls, d):
        variable = cls(d["name"])
        variable.value = d["value"]
        variable.start = d["start"]
        variable.fixed = d["fixed"]

        return variable


DelayArgument = namedtuple("DelayArgument", ["expr", "duration"])


class Model:
    def __init__(self):
        self.states = []
        self.der_states = []
        self.alg_states = []
        self.inputs = []
        self.outputs = []
        self.constants = []
        self.parameters = []
        self.string_constants = []
        self.string_parameters = []
        self.equations = []
        self.initial_equations = []
        self.time = ca.MX.sym("time")
        self.delay_states = []
        self.delay_arguments = []
        self.alias_relation = AliasRelation()
        self._expand_mx_func = lambda x: x

    def __str__(self):
        r = ""
        r += "Model\n"
        r += "time: " + str(self.time) + "\n"
        r += "states: " + str(self.states) + "\n"
        r += "der_states: " + str(self.der_states) + "\n"
        r += "alg_states: " + str(self.alg_states) + "\n"
        r += "inputs: " + str(self.inputs) + "\n"
        r += "outputs: " + str(self.outputs) + "\n"
        r += "constants: " + str(self.constants) + "\n"
        r += "parameters: " + str(self.parameters) + "\n"
        r += "string_constants: " + str(self.string_constants) + "\n"
        r += "string_parameters: " + str(self.string_parameters) + "\n"
        r += "equations: " + str(self.equations) + "\n"
        r += "initial equations: " + str(self.initial_equations) + "\n"
        return r

    def check_balanced(self):
        n_states = sum(
            v.symbol.size1() * v.symbol.size2()
            for v in itertools.chain(self.states, self.alg_states)
        )
        n_equations = sum(e.size1() * e.size2() for e in self.dae_residual_function.mx_out())
        if n_states == n_equations:
            logger.info("System is balanced.")
        else:
            logger.warning(
                "System is not balanced.  "
                "Number of states is {}, number of equations is {}.".format(n_states, n_equations)
            )

    def _post_checks(self):
        # We do not support delayMax yet, so delay durations can only depend
        # on constants, parameters and fixed inputs.
        if self.delay_states:
            delay_durations = ca.veccat(*(x.duration for x in self.delay_arguments))
            disallowed_duration_symbols = ca.vertcat(
                self.time,
                ca.veccat(*self._symbols(self.states)),
                ca.veccat(*self._symbols(self.der_states)),
                ca.veccat(*self._symbols(self.alg_states)),
                ca.veccat(*(x.symbol for x in self.inputs if not x.fixed)),
            )

            if ca.depends_on(delay_durations, disallowed_duration_symbols):
                raise ValueError(
                    "Delay durations can only depend on parameters, constants and fixed inputs."
                )

    @staticmethod
    def _symbols(variables):
        return [v.symbol for v in variables]

    def _substitute_delay_arguments(self, delay_arguments, symbols, values):
        exprs = ca.substitute(
            [ca.MX(argument.expr) for argument in delay_arguments], symbols, values
        )
        durations = ca.substitute(
            [ca.MX(argument.duration) for argument in delay_arguments], symbols, values
        )
        return [DelayArgument(expr, duration) for expr, duration in zip(exprs, durations)]

    @staticmethod
    def _expand_simplify_mx(equations):
        # Sometimes MX expressions can end up horribly nested and complicated,
        # which makes further simplifications like alias detection difficult.
        # We can simplify the equations by expanding to SX, and then
        # rebuilding them with MX symbols again.

        if not equations:
            return []

        symbols_mx = ca.symvar(ca.veccat(*equations))
        assert all(
            x.shape == (1, 1) for x in symbols_mx
        ), "Vector/Matrix SX symbols cannot be mapped"

        f_mx = ca.Function("tmp_mx", symbols_mx, equations).expand()
        symbols_sx = [ca.SX.sym(x.name(), *x.shape) for x in symbols_mx]
        sx_equations = f_mx.call(symbols_sx)

        sx_to_mx_map = dict(zip(symbols_sx, symbols_mx))

        def _sx_to_mx(sx_expr):
            if not sx_expr.is_scalar():
                rows = []
                for i in range(sx_expr.size1()):
                    cols = []
                    for j in range(sx_expr.size2()):
                        cols.append(_sx_to_mx(sx_expr[i, j]))
                    rows.append(cols)
                return ca.vertcat(*[ca.horzcat(*row) for row in rows])
            elif sx_expr.op() == ca.OP_PARAMETER:
                assert sx_expr in sx_to_mx_map.keys()
                return sx_to_mx_map[sx_expr]
            elif sx_expr.op() == ca.OP_CONST:
                return ca.MX(ca.DM(sx_expr))
            elif sx_expr.n_dep() == 1:
                return ca.MX.unary(sx_expr.op(), _sx_to_mx(sx_expr.dep(0)))
            elif sx_expr.n_dep() == 2:
                return ca.MX.binary(
                    sx_expr.op(), _sx_to_mx(sx_expr.dep(0)), _sx_to_mx(sx_expr.dep(1))
                )
            else:
                raise Exception("Unsupported operation in SX expression.")

        equations = []
        for eq in sx_equations:
            try:
                equations.append(_sx_to_mx(eq))
            except Exception:
                equations.append(eq)

        return equations

    def _substitute_metadata(self, symbols, values):
        substitutions = []
        for variable in itertools.chain(
            self.states, self.alg_states, self.inputs, self.parameters, self.constants
        ):
            for attribute in CASADI_ATTRIBUTES:
                value = getattr(variable, attribute)
                if isinstance(value, ca.MX) and not value.is_constant():
                    substitutions.append((value, variable, attribute))

        if len(substitutions) == 0:
            return

        expressions, variables, attributes = zip(*substitutions)
        expressions = ca.substitute(expressions, symbols, values)

        for variable, attribute, value in zip(variables, attributes, expressions):
            if variable.python_type in {int, float}:
                if isinstance(value, ca.MX) and value.is_constant() and value.shape == (1, 1):
                    if attribute in {"value", "start", "min", "max", "nominal"}:
                        # Inf is common/the default for min/max. Because
                        # integers cannot represent this value, we cast them
                        # to floats instead.
                        if value.is_regular():
                            value = variable.python_type(float(value))
                        else:
                            value = float(value)

            setattr(variable, attribute, value)

    def _expand_vectors(self):
        logger.info("Expanding vectors")

        symbols = []
        values = []

        for vgroup in [
            "states",
            "der_states",
            "alg_states",
            "inputs",
            "parameters",
            "constants",
        ]:
            old_vars = getattr(self, vgroup)
            new_vars = []
            for old_var in old_vars:
                # For delayed states we do not have any reliable shape
                # information available due to it being an arbitrary
                # expression, so we just always expand.
                if (
                    set(old_var.symbol._modelica_shape) != {(None,)}
                    or old_var.symbol.name() in self.delay_states
                ):
                    expanded_symbols = []

                    # Prepare a component_name_format in which the different possible
                    # combinations of indices can later be inserted.
                    # For example: a.b[{}].c[{},{}]
                    if old_var.symbol.name() in self.delay_states:
                        # For delay states we use the _modelica_shape as is
                        iterator_shape = old_var.symbol._modelica_shape
                        component_name_format = (
                            old_var.symbol.name()
                            + "["
                            + ",".join("{}" for _ in range(len(iterator_shape)))
                            + "]"
                        )
                    else:
                        # For (nested) array symbols the _modelica_shape contains a tuple of
                        # tuples containing the shape for each of the nested symbols.
                        # For example, symbol name a.b.c and shape ((None,),(3,),(1,3)) means
                        # a was a scalar, b a 1d array of size [3] and c a 2d array of size
                        # [1,3].
                        modelica_shape = old_var.symbol._modelica_shape
                        old_sym_name = old_var.symbol.name()

                        # Expanded derivative symbols get their indices inside the parantheses.
                        # For a symbol called "der(a.x)", the prefix will be "der(", the postfix
                        # will be ")", with old_sym_name becoming "a.x".
                        # For a symbol called "a.x", the prefix and postfix will be empty strings.
                        prefix, old_sym_name, postfix = re.match(
                            r"((?:der\()*|\b)(.*?)([\)]*|\b)$", old_sym_name
                        ).groups()

                        symbol_names = old_sym_name.split(".")
                        assert len(symbol_names) == len(modelica_shape)
                        component_name_format = ""
                        for i, symbol_name in enumerate(symbol_names):
                            if i != 0:
                                component_name_format += "."
                            component_name_format += symbol_name
                            if modelica_shape[i] != (None,):
                                component_name_format += (
                                    "["
                                    + ",".join("{}" for _ in range(len(modelica_shape[i])))
                                    + "]"
                                )

                        component_name_format = prefix + component_name_format + postfix

                        iterator_shape = tuple(
                            [d for var_shape in modelica_shape for d in var_shape if d is not None]
                        )

                    # Generate symbols for each possible combination of indices
                    for ind in np.ndindex(iterator_shape):
                        component_symbol = ca.MX.sym(
                            component_name_format.format(*tuple(i + 1 for i in ind))
                        )
                        component_var = Variable(component_symbol, old_var.python_type)
                        for attribute in CASADI_ATTRIBUTES:
                            # Can't convert 3D arrays to MX, so we convert to nparray instead
                            value = getattr(old_var, attribute)
                            if not isinstance(value, ca.MX):
                                if np.isscalar(value):
                                    # Just assign without indexing
                                    val = value
                                elif isinstance(value, list):
                                    val = value
                                    for i in ind:
                                        val = val[i]
                                elif isinstance(value, (ca.DM, np.ndarray)):
                                    val = value[ind]
                                    if old_var.python_type in {float, int}:
                                        val = old_var.python_type(val)
                                else:
                                    raise TypeError("Unexpected type from CasADi generator")
                            else:
                                if np.prod(value.shape) == 1:
                                    # Just assign without indexing
                                    val = value
                                else:
                                    val = value[ind]

                            setattr(component_var, attribute, val)

                        expanded_symbols.append(component_var)

                    s = (
                        old_var.symbol._mx
                        if isinstance(old_var.symbol, _MTensor)
                        else old_var.symbol
                    )
                    symbols.append(s)
                    values.append(
                        ca.reshape(
                            ca.vertcat(*[x.symbol for x in expanded_symbols]),
                            *tuple(reversed(s.shape)),
                        ).T
                    )
                    new_vars.extend(expanded_symbols)

                    # Replace variable in delay expressions and durations if needed
                    try:
                        assert len(self.delay_states) == len(self.delay_arguments)

                        i = self.delay_states.index(old_var.symbol.name())
                    except ValueError:
                        pass
                    else:
                        delay_state = self.delay_states.pop(i)
                        delay_argument = self.delay_arguments.pop(i)

                        for ind in np.ndindex(old_var.symbol._modelica_shape):
                            new_name = "{}[{}]".format(
                                delay_state, ",".join(str(i + 1) for i in ind)
                            )

                            self.delay_states.append(new_name)
                            self.delay_arguments.append(
                                DelayArgument(delay_argument.expr[ind], delay_argument.duration)
                            )

                    # Replace variable in list of outputs if needed
                    try:
                        i = self.outputs.index(old_var.symbol.name())
                    except ValueError:
                        pass
                    else:
                        self.outputs.pop(i)
                        for new_s in reversed(expanded_symbols):
                            self.outputs.insert(i, new_s.symbol.name())
                else:
                    new_vars.append(old_var)

            setattr(self, vgroup, new_vars)

        if len(self.equations) > 0:
            self.equations = ca.substitute(self.equations, symbols, values)
            self.equations = list(
                itertools.chain.from_iterable(ca.vertsplit(ca.vec(eq)) for eq in self.equations)
            )
        if len(self.initial_equations) > 0:
            self.initial_equations = ca.substitute(self.initial_equations, symbols, values)
            self.initial_equations = list(
                itertools.chain.from_iterable(
                    ca.vertsplit(ca.vec(eq)) for eq in self.initial_equations
                )
            )
        if len(self.delay_arguments) > 0:
            self.delay_arguments = self._substitute_delay_arguments(
                self.delay_arguments, symbols, values
            )

        # Make sure that the naming in the main loop and the delay argument loop match
        input_names = [v.symbol.name() for v in self.inputs]
        assert set(self.delay_states).issubset(input_names)

        # Replace values in metadata
        self._substitute_metadata(symbols, values)

    def simplify(self, options):
        options = _merge_default_options(options)

        alg_states_left = 0

        for simplification_iter in range(SIMPLIFICATION_LOOP_LIMIT):
            if options.get("iterative_simplification", False):
                logger.info("Simplification iteration {}".format(simplification_iter + 1))

            self._simplify_once(options)

            if options.get("iterative_simplification", False) and alg_states_left != len(
                self.alg_states
            ):
                alg_states_left = len(self.alg_states)
                simplification_iter += 1
            else:
                break
        else:
            logger.warning("Simplification exceeded maximum iteration limit.")

        logger.info("Finished model simplification")

    def _simplify_once(self, options):
        if options["expand_vectors"] and options["expand_mx"]:
            # If we are expanding MX to SX, we do the expansion of
            # vectors first, such that we are be able to detect more aliases
            # (e.g individual array elements, or elements in a for loop).
            self._expand_vectors()

            # Expand to SX, and back again to MX such that the remaining
            # simplification steps can be applied. We also do this to make
            # sure we only ever expose MX to the outside world.
            self.equations = self._expand_simplify_mx(self.equations)
            self.initial_equations = self._expand_simplify_mx(self.initial_equations)

        if options["resolve_parameter_values"]:
            # Resolve all numeric parameter values that are set (i.e. value
            # not equal to NaN). Inline the values, but keep the parameter
            # Variables around. Note that the associated parameter MX symbols
            # then do not appear in any min/max/nominal/value metadata (e.g.
            # those of other parameters).
            current_parameters_and_constants = {*self.parameters, *self.constants}

            for _ in range(SUBSTITUTE_LOOP_LIMIT):
                symbols, values = [], []
                next_parameters_and_constants = []

                for p in current_parameters_and_constants:
                    if isinstance(p.value, (list, np.ndarray)):
                        p.value = np.array(p.value)

                        if not np.issubdtype(p.value.dtype, np.number):
                            raise NotImplementedError(
                                "Only parameters arrays with numeric values can be simplified"
                            )

                        if not np.any(np.isnan(p.value)):
                            symbols.append(p.symbol)
                            values.append(p.value)
                    else:
                        value = ca.MX(p.value)
                        if value.is_constant() and value.is_regular():
                            symbols.append(p.symbol)
                            values.append(value)
                        else:
                            next_parameters_and_constants.append(p)

                if not values:
                    break

                self._substitute_metadata(symbols, values)

                current_parameters_and_constants = next_parameters_and_constants

        if options["replace_parameter_expressions"]:
            logger.info("Replacing parameter expressions")

            simple_parameters, symbols, values = [], [], []
            for p in self.parameters:
                if isinstance(p.value, list):
                    p.value = np.array(p.value)

                    if not np.issubdtype(p.value.dtype, np.number):
                        raise NotImplementedError(
                            "Only parameters arrays with numeric values can be simplified"
                        )

                    simple_parameters.append(p)
                else:
                    value = ca.MX(p.value)
                    if value.is_constant():
                        simple_parameters.append(p)
                    else:
                        symbols.append(p.symbol)
                        values.append(value)

            self.parameters = simple_parameters

            if len(values) > 0:
                # Resolve expressions that include other, non-simple parameter
                # expressions.
                for _ in range(SUBSTITUTE_LOOP_LIMIT):
                    new_values = ca.substitute(values, symbols, values)
                    converged = ca.is_equal(
                        ca.veccat(*values), ca.veccat(*new_values), CASADI_COMPARISON_DEPTH
                    )
                    values = new_values
                    if converged:
                        break
                else:
                    logger.warning("Substitution of expressions exceeded maximum iteration limit.")

                if len(self.equations) > 0:
                    self.equations = ca.substitute(self.equations, symbols, values)
                if len(self.initial_equations) > 0:
                    self.initial_equations = ca.substitute(self.initial_equations, symbols, values)
                if len(self.delay_arguments) > 0:
                    self.delay_arguments = self._substitute_delay_arguments(
                        self.delay_arguments, symbols, values
                    )

                # Replace parameter expressions in metadata
                self._substitute_metadata(symbols, values)

        if options["replace_constant_expressions"]:
            logger.info("Replacing constant expressions")

            simple_constants, symbols, values = [], [], []
            for c in self.constants:
                value = ca.MX(c.value)
                if value.is_constant():
                    simple_constants.append(c)
                else:
                    symbols.append(c.symbol)
                    values.append(value)

            self.constants = simple_constants

            if len(values) > 0:
                # Resolve expressions that include other, non-simple parameter
                # expressions.
                for _ in range(SUBSTITUTE_LOOP_LIMIT):
                    new_values = ca.substitute(values, symbols, values)
                    converged = ca.is_equal(
                        ca.veccat(*values), ca.veccat(*new_values), CASADI_COMPARISON_DEPTH
                    )
                    values = new_values
                    if converged:
                        break
                else:
                    logger.warning("Substitution of expressions exceeded maximum iteration limit.")

                if len(self.equations) > 0:
                    self.equations = ca.substitute(self.equations, symbols, values)
                if len(self.initial_equations) > 0:
                    self.initial_equations = ca.substitute(self.initial_equations, symbols, values)
                if len(self.delay_arguments) > 0:
                    self.delay_arguments = self._substitute_delay_arguments(
                        self.delay_arguments, symbols, values
                    )

                # Replace constant expressions in metadata
                self._substitute_metadata(symbols, values)

        if options["eliminate_constant_assignments"]:
            logger.info("Elimating constant variable assignments")

            alg_states = OrderedDict([(s.symbol.name(), s) for s in self.alg_states])

            reduced_equations = []
            for eq in self.equations:
                if eq.is_symbolic() and eq.name() in alg_states:
                    constant = alg_states.pop(eq.name())
                    constant.value = 0.0

                    self.constants.append(constant)

                    # Skip this equation
                    continue

                if eq.n_dep() == 2 and (eq.is_op(ca.OP_SUB) or eq.is_op(ca.OP_ADD)):
                    if (
                        eq.dep(0).is_symbolic()
                        and eq.dep(0).name() in alg_states
                        and eq.dep(1).is_constant()
                    ):
                        variable = eq.dep(0)
                        value = eq.dep(1)
                    elif (
                        eq.dep(1).is_symbolic()
                        and eq.dep(1).name() in alg_states
                        and eq.dep(0).is_constant()
                    ):
                        variable = eq.dep(1)
                        value = eq.dep(0)
                    else:
                        variable = None
                        value = None

                    if variable is not None:
                        constant = alg_states.pop(variable.name())

                        if eq.is_op(ca.OP_SUB):
                            constant.value = value
                        else:
                            constant.value = -value

                        self.constants.append(constant)

                        # Skip this equation
                        continue

                # Keep this equation
                reduced_equations.append(eq)

            # Eliminate alias variables
            self.alg_states = list(alg_states.values())
            self.equations = reduced_equations

        if options["replace_parameter_values"]:
            logger.info("Replacing parameter values")

            # N.B. Any parameter expression elimination must be done first.
            unspecified_parameters, symbols, values = [], [], []
            for p in self.parameters:
                if ca.MX(p.value).is_constant() and ca.MX(p.value).is_regular():
                    symbols.append(p.symbol)
                    values.append(p.value)
                else:
                    unspecified_parameters.append(p)

            if len(self.equations) > 0:
                self.equations = ca.substitute(self.equations, symbols, values)
            if len(self.initial_equations) > 0:
                self.initial_equations = ca.substitute(self.initial_equations, symbols, values)
            self.parameters = unspecified_parameters

            # Replace parameter values in metadata
            self._substitute_metadata(symbols, values)

        if options["replace_constant_values"]:
            logger.info("Replacing constant values")

            # N.B. Any parameter expression elimination must be done first.
            symbols = self._symbols(self.constants)
            values = [v.value for v in self.constants]
            if len(self.equations) > 0:
                self.equations = ca.substitute(self.equations, symbols, values)
            if len(self.initial_equations) > 0:
                self.initial_equations = ca.substitute(self.initial_equations, symbols, values)
            if len(self.delay_arguments) > 0:
                self.delay_arguments = self._substitute_delay_arguments(
                    self.delay_arguments, symbols, values
                )

            # Also remove the aliases of any of the constants from the alias relation
            for constant in self.constants:
                if constant.aliases:
                    # Note that we will only get here if simplify is called multiple times on the
                    # same model, as alias detection is done _after_ replacing constant values.
                    # Otherwise a variable will not have any aliases yet.
                    self.alias_relation.remove(constant.symbol.name())

            self.constants = []

            # Replace constant values in metadata
            self._substitute_metadata(symbols, values)

        if options["eliminable_variable_expression"] is not None:
            logger.info(
                "Elimating variables that match the regular expression {}".format(
                    options["eliminable_variable_expression"]
                )
            )

            # Due to CasADi's ca.is_equal not properly matching short-
            # circuited MX if-else expressions, we can end up in an infinite
            # loop. With expanded (to SX and back) equations we are safe, as
            # SX does not support short-circuiting.
            if not options["expand_mx"]:
                raise Exception(
                    "The use of `eliminable_variable_expression` requires `expand_mx` set to True"
                )

            p = re.compile(options["eliminable_variable_expression"])

            states = OrderedDict([(s.symbol.name(), s) for s in self.states])
            alg_states = OrderedDict([(s.symbol.name(), s) for s in self.alg_states])

            all_states = OrderedDict(
                [(s.symbol.name(), s) for s in itertools.chain(self.states, self.alg_states)]
            )

            # NOTE: dictionary keys are the correspondig _state_'s name, not
            # the names of the derivative states themselves.
            der_states = OrderedDict(zip(states.keys(), self.der_states))

            def extract_assignment(eq):
                if eq.is_symbolic() and eq.name() in all_states and p.match(eq.name()):
                    return eq, 0.0

                if eq.is_op(ca.OP_IF_ELSE_ZERO):
                    cond = eq.dep(0)
                    nested_eq = eq.dep(1)

                    variable, value = extract_assignment(nested_eq)
                    if variable is not None:
                        return variable, ca.if_else(cond, value, 0)

                if eq.n_dep() == 2 and (eq.is_op(ca.OP_SUB) or eq.is_op(ca.OP_ADD)):
                    # Prefer eliminating alg states
                    if (
                        eq.dep(0).is_symbolic()
                        and eq.dep(0).name() in alg_states
                        and p.match(eq.dep(0).name())
                    ):
                        variable = eq.dep(0)
                        value = eq.dep(1)
                    elif (
                        eq.dep(1).is_symbolic()
                        and eq.dep(1).name() in alg_states
                        and p.match(eq.dep(1).name())
                    ):
                        variable = eq.dep(1)
                        value = eq.dep(0)
                    elif (
                        eq.dep(0).is_symbolic()
                        and eq.dep(0).name() in states
                        and p.match(eq.dep(0).name())
                    ):
                        variable = eq.dep(0)
                        value = eq.dep(1)
                    elif (
                        eq.dep(1).is_symbolic()
                        and eq.dep(1).name() in states
                        and p.match(eq.dep(1).name())
                    ):
                        variable = eq.dep(1)
                        value = eq.dep(0)
                    else:
                        variable = None
                        value = None

                    if variable is not None:
                        if eq.is_op(ca.OP_SUB):
                            return variable, value
                        else:
                            return variable, -value

                if (
                    eq.n_dep() == 2
                    and eq.is_op(ca.OP_ADD)
                    and eq.dep(0).is_op(ca.OP_IF_ELSE_ZERO)
                    and eq.dep(1).is_op(ca.OP_IF_ELSE_ZERO)
                ):
                    # Once passed through _expand_simplify_mx, a CasADi if_else
                    # is converted to a sum of two if_else_zero operations.

                    cond_1 = eq.dep(0).dep(0)
                    cond_2 = eq.dep(1).dep(0)
                    nested_eq_1 = eq.dep(0).dep(1)
                    nested_eq_2 = eq.dep(1).dep(1)

                    variable_1, value_1 = extract_assignment(nested_eq_1)
                    variable_2, value_2 = extract_assignment(nested_eq_2)

                    if (
                        variable_1 is not None
                        and variable_2 is not None
                        and variable_1.name() == variable_2.name()
                    ):
                        return variable_1, ca.if_else(cond_1, value_1, 0) + ca.if_else(
                            cond_2, value_2, 0
                        )

                # Nothing found
                return None, None

            variables = []
            values = []

            def get_derivative(expr):
                if expr.is_constant():
                    return 0.0
                elif expr.is_symbolic():
                    if expr.name() in states:
                        return der_states[expr.name()].symbol
                    elif expr.name() in alg_states:
                        # This algebraic state must now become a differentiated state.
                        states[expr.name()] = alg_states.pop(expr.name())
                        der_sym = ca.MX.sym("der({})".format(expr.name()))
                        der_states[expr.name()] = Variable(der_sym, float)
                        return der_sym
                    else:
                        return 0.0
                else:
                    # Differentiate using CasADi and chain rule
                    orig_deps = ca.symvar(expr)
                    deps = ca.vertcat(*orig_deps)
                    J = ca.Function("J", [deps], [ca.jacobian(expr, deps)])
                    J_sparsity = J.sparsity_out(0)
                    der_deps = [
                        get_derivative(dep) if J_sparsity.has_nz(0, j) else ca.DM.zeros(dep.size())
                        for j, dep in enumerate(orig_deps)
                    ]
                    return ca.mtimes(J(deps), ca.vertcat(*der_deps))

            reduced_equations = []
            for eq in self.equations:
                variable, value = extract_assignment(eq)
                if variable is not None:
                    if variable.name() in states:
                        # Mark derivative state for replacement too.
                        derivative = get_derivative(value)

                        del states[variable.name()]

                        variables.append(der_states.pop(variable.name()).symbol)
                        values.append(derivative)

                    elif variable.name() in alg_states:
                        del alg_states[variable.name()]

                    variables.append(variable)
                    values.append(value)

                    # Skip this equation
                    continue

                # Keep this equation
                reduced_equations.append(eq)

            # Eliminate alias variables
            self.states = list(states.values())
            self.alg_states = list(alg_states.values())
            self.der_states = list(der_states.values())

            self.equations = reduced_equations

            if len(values) > 0:
                # Resolve expressions that include other, non-simple parameter
                # expressions.
                for _ in range(SUBSTITUTE_LOOP_LIMIT):
                    new_values = ca.substitute(values, variables, values)
                    converged = ca.is_equal(
                        ca.veccat(*values), ca.veccat(*new_values), CASADI_COMPARISON_DEPTH
                    )
                    values = new_values
                    if converged:
                        break
                else:
                    logger.warning("Substitution of expressions exceeded maximum iteration limit.")

                if len(self.equations) > 0:
                    self.equations = ca.substitute(self.equations, variables, values)
                if len(self.initial_equations) > 0:
                    self.initial_equations = ca.substitute(
                        self.initial_equations, variables, values
                    )
                if len(self.delay_arguments) > 0:
                    self.delay_arguments = self._substitute_delay_arguments(
                        self.delay_arguments, variables, values
                    )

        if options["expand_vectors"] and not options["expand_mx"]:
            # If we are _not_ expanding MX to SX, we do the expansion of vectors here
            self._expand_vectors()

        if options["factor_and_simplify_equations"]:
            # Operations that preserve the equivalence of an equation
            # TODO: There may be more, but this is the most frequent set
            unary_ops = ca.OP_NEG, ca.OP_FABS, ca.OP_SQRT
            binary_ops = ca.OP_MUL, ca.OP_DIV

            # Recursive factor and simplify function
            def factor_and_simplify(eq):
                # These are ops that can simply be dropped
                if eq.n_dep() == 1 and eq.op() in unary_ops:
                    return factor_and_simplify(eq.dep())

                # These are binary ops and can get a little tricky
                # For now, we just drop constant divisors or multipliers
                elif eq.n_dep() == 2 and eq.op() in binary_ops:
                    if eq.dep(1).is_constant():
                        return factor_and_simplify(eq.dep(0))
                    elif eq.dep(0).is_constant() and eq.op() == ca.OP_MUL:
                        return factor_and_simplify(eq.dep(1))

                # If no hits, return unmodified
                return eq

            # Do the simplifications
            simplified_equations = [factor_and_simplify(eq) for eq in self.equations]

            # Debugging output
            if logger.getEffectiveLevel() == logging.DEBUG:
                changed_equations = [
                    (o, s) for o, s in zip(self.equations, simplified_equations) if o is not s
                ]
                for orig, simp in changed_equations:
                    logger.debug("Equation {} simplified to {}".format(orig, simp))

            # Store changes
            self.equations = simplified_equations

        if options["detect_aliases"]:
            logger.info("Detecting aliases")

            states = OrderedDict([(s.symbol.name(), s) for s in self.states])
            der_states = OrderedDict([(s.symbol.name(), s) for s in self.der_states])
            alg_states = OrderedDict([(s.symbol.name(), s) for s in self.alg_states])
            inputs = OrderedDict([(s.symbol.name(), s) for s in self.inputs])
            parameters = OrderedDict([(s.symbol.name(), s) for s in self.parameters])
            constants = OrderedDict([(s.symbol.name(), s) for s in self.constants])

            all_states = OrderedDict()
            all_states.update(states)
            all_states.update(der_states)
            all_states.update(alg_states)
            all_states.update(inputs)
            all_states.update(parameters)
            all_states.update(constants)

            # For now, we only eliminate algebraic states.
            do_not_eliminate = set(
                list(der_states) + list(states) + list(inputs) + list(parameters) + list(constants)
            )

            # When doing a `detect_aliases` pass multiple times, we have to avoid
            # handling aliases we already handled before as their states no longer
            # exist. To know which alias is new and which is old, we make a copy here
            # of the alias relation.
            old_alias_relation = self.alias_relation.copy()

            def _make_alias(deps, negative_alias=False):
                """
                Tries to alias the deps to each other. Returns True if this
                was possible and done, False otherwise.
                """
                if deps[0].name() in alg_states:
                    alg_state = deps[0]
                    other_state = deps[1]
                elif deps[1].name() in alg_states:
                    alg_state = deps[1]
                    other_state = deps[0]
                else:
                    alg_state = None
                    other_state = None

                # If both states are algebraic, we need to decide which to eliminate
                if deps[0].name() in alg_states and deps[1].name() in alg_states:
                    # Most of the time it does not matter which one we eliminate.
                    # The exception is if alg_state has already been aliased to a
                    # variable in do_not_eliminate. If this is the case, setting the
                    # states in the default order will cause the new canonical variable
                    # to be other_state, unseating (and eliminating) the current
                    # canonical variable (which is in do_not_eliminate).
                    if (
                        self.alias_relation.canonical_signed(alg_state.name())[0]
                        in do_not_eliminate
                    ):
                        # swap the states
                        other_state, alg_state = alg_state, other_state

                if alg_state is not None:
                    # If either state is a derivative state, and aliasing of those
                    # is not allowed, skip aliasing them
                    if not options["allow_derivative_aliases"] and (
                        alg_state.name() in der_states or other_state.name() in der_states
                    ):
                        return False

                    # Check to see if we are linking two entries in do_not_eliminate
                    if (
                        self.alias_relation.canonical_signed(alg_state.name())[0]
                        in do_not_eliminate
                        and self.alias_relation.canonical_signed(other_state.name())[0]
                        in do_not_eliminate
                    ):
                        # Don't do anything for now, we only eliminate alg_states
                        pass

                    else:
                        # Eliminate alg_state by aliasing it to other_state
                        if negative_alias:
                            self.alias_relation.add(other_state.name(), "-" + alg_state.name())
                        else:
                            self.alias_relation.add(other_state.name(), alg_state.name())

                        # To keep equations balanced, drop this equation
                        return True

                return False

            def _detect_alias(eq):
                # We do fast checks first, and the slower (but more generic)
                # checks after.
                deps = ca.symvar(eq)

                if len(deps) == 2:
                    if eq.n_dep() == 2 and (eq.is_op(ca.OP_SUB) or eq.is_op(ca.OP_ADD)):
                        if eq.dep(0).is_symbolic() and eq.dep(1).is_symbolic():
                            return deps, eq.is_op(ca.OP_ADD)

                deps_names = [x.name() for x in deps]
                non_param_deps = [
                    sym
                    for name, sym in zip(deps_names, deps)
                    if name not in parameters and name not in constants
                ]

                for d in [deps, non_param_deps]:
                    if not len(d) == 2:
                        continue

                    # Check with substitute, which is a more expensive operation
                    if ca.substitute(eq, d[0], d[1]).is_zero():
                        return d, False

                    elif ca.substitute(eq, d[0], -1 * d[1]).is_zero():
                        return d, True

                return [], False

            reduced_equations = []
            for eq in self.equations:
                if options["expand_vectors"] and not options["expand_mx"]:
                    # Equation might have many "shadow" dependencies due to
                    # vector/array expansion. By using .expand() and SX
                    # symbols for the evaluation, we can figure out what the
                    # real dependencies are.
                    s_mx = ca.symvar(eq)
                    assert all(
                        x.shape == (1, 1) for x in s_mx
                    ), "Vector/Matrix SX symbols cannot be mapped"
                    f = ca.Function("tmp", s_mx, [eq]).expand()
                    s_sx = [ca.SX.sym(x.name(), *x.shape) for x in s_mx]
                    eq_sx = f.call(s_sx)[0]

                    symbols, negative_alias = _detect_alias(eq_sx)

                    if symbols:
                        deps_map = dict(zip(s_sx, s_mx))
                        symbols = [deps_map[s] for s in symbols]
                else:
                    # We can just use the MX equation directly
                    symbols, negative_alias = _detect_alias(eq)

                if symbols and _make_alias(symbols, negative_alias):
                    continue
                else:
                    # Keep this equation
                    reduced_equations.append(eq)

            # Eliminate alias variables
            variables, values = [], []
            for canonical, aliases in self.alias_relation:
                canonical_state = all_states[canonical]

                python_type = canonical_state.python_type
                start = canonical_state.start
                m, M = canonical_state.min, canonical_state.max
                nominal = canonical_state.nominal
                fixed = canonical_state.fixed

                for alias in aliases:
                    if (
                        len(old_alias_relation.aliases(alias)) > 1
                        and alias not in old_alias_relation.canonical_variables
                    ):
                        # We already handled this alias in a previous pass of `detect_aliases`
                        continue

                    if alias[0] == "-":
                        sign = -1
                        alias = alias[1:]
                    else:
                        sign = 1

                    alias_state = all_states[alias]

                    variables.append(alias_state.symbol)
                    values.append(sign * canonical_state.symbol)

                    # If any of the aliases has a nonstandard type, apply it to
                    # the canonical state as well
                    if alias_state.python_type != float:
                        python_type = alias_state.python_type

                    # If any of the aliases has a nondefault start value, apply it to
                    # the canonical state as well
                    if not isinstance(alias_state.start, _DefaultValue):
                        alias_start_mx = ca.MX(alias_state.start)
                        if not isinstance(start, _DefaultValue):
                            start_mx = ca.MX(start)
                            # If the state already has a non-default start
                            # attribute we check for conflicts.
                            if (
                                start_mx.is_constant() != ca.MX(alias_start_mx).is_constant()
                                or (
                                    start_mx.is_symbolic()
                                    and str(start_mx) != str(sign * alias_start_mx)
                                )
                                or start != alias_start_mx
                            ):
                                logger.warning(
                                    "Current start attribute of canonical variable '{}' ({})"
                                    " conflicts with that of its alias '{}' ({})."
                                    " Will keep existing value of {}.".format(
                                        canonical, start, alias, alias_start_mx, start
                                    )
                                )
                            else:
                                # Alias has equal start attribute, so nothing to do
                                pass
                        else:
                            start = sign * alias_state.start

                    # The intersection of all bound ranges applies
                    m = ca.fmax(m, alias_state.min if sign == 1 else -alias_state.max)
                    M = ca.fmin(M, alias_state.max if sign == 1 else -alias_state.min)

                    # Take the largest nominal of all aliases
                    nominal = ca.fmax(nominal, alias_state.nominal)

                    # If any of the aliases is fixed, the canonical state is as well
                    fixed = ca.fmax(fixed, alias_state.fixed)

                    del all_states[alias]

                assert isinstance(aliases, set)
                canonical_state.aliases = aliases
                canonical_state.python_type = python_type
                canonical_state.start = start
                canonical_state.min = m
                canonical_state.max = M
                canonical_state.nominal = nominal
                canonical_state.fixed = fixed

            self.states = [v for k, v in all_states.items() if k in states]
            self.der_states = [v for k, v in all_states.items() if k in der_states]
            self.alg_states = [v for k, v in all_states.items() if k in alg_states]
            self.inputs = [v for k, v in all_states.items() if k in inputs]
            self.parameters = [v for k, v in all_states.items() if k in parameters]
            self.equations = reduced_equations

            if len(self.equations) > 0:
                self.equations = ca.substitute(self.equations, variables, values)
            if len(self.initial_equations) > 0:
                self.initial_equations = ca.substitute(self.initial_equations, variables, values)
            if len(self.delay_arguments) > 0:
                self.delay_arguments = self._substitute_delay_arguments(
                    self.delay_arguments, variables, values
                )

        if options["reduce_affine_expression"]:
            logger.info("Collapsing model into an affine expression")

            for equation_list in ["equations", "initial_equations"]:
                equations = getattr(self, equation_list)
                if len(equations) > 0:
                    states = ca.veccat(
                        *self._symbols(
                            itertools.chain(
                                self.states, self.der_states, self.alg_states, self.inputs
                            )
                        )
                    )
                    constants = ca.veccat(*self._symbols(self.constants))
                    parameters = ca.veccat(*self._symbols(self.parameters))

                    equations = ca.veccat(*equations)

                    Af = ca.Function(
                        "Af", [states, constants, parameters], [ca.jacobian(equations, states)]
                    )
                    bf = ca.Function("bf", [states, constants, parameters], [equations])

                    # Work around CasADi issue #172
                    if len(self.constants) == 0 or not ca.depends_on(equations, constants):
                        constants = 0
                    else:
                        logger.warning(
                            "Not all constants have been eliminated.  As a result, the affine DAE expression will use a symbolic matrix, as opposed to a numerical sparse matrix."
                        )
                    if len(self.parameters) == 0 or not ca.depends_on(equations, parameters):
                        parameters = 0
                    else:
                        logger.warning(
                            "Not all parameters have been eliminated.  As a result, the affine DAE expression will use a symbolic matrix, as opposed to a numerical sparse matrix."
                        )

                    A = Af(0, constants, parameters)
                    b = bf(0, constants, parameters)

                    # Replace veccat'ed states with brand new state vectors so as to avoid the value copy operations induced by veccat.
                    self._states_vector = ca.MX.sym(
                        "states_vector", sum([s.numel() for s in self._symbols(self.states)])
                    )
                    self._der_states_vector = ca.MX.sym(
                        "der_states_vector",
                        sum([s.numel() for s in self._symbols(self.der_states)]),
                    )
                    self._alg_states_vector = ca.MX.sym(
                        "alg_states_vector",
                        sum([s.numel() for s in self._symbols(self.alg_states)]),
                    )
                    self._inputs_vector = ca.MX.sym(
                        "inputs_vector", sum([s.numel() for s in self._symbols(self.inputs)])
                    )

                    states_vector = ca.vertcat(
                        self._states_vector,
                        self._der_states_vector,
                        self._alg_states_vector,
                        self._inputs_vector,
                    )
                    equations = [ca.reshape(ca.mtimes(A, states_vector), equations.shape) + b]
                    setattr(self, equation_list, equations)

        if options["expand_mx"]:
            logger.info("Expanded MX functions will be returned")
            self._expand_mx_func = lambda x: x.expand()

        logger.info("Finished model simplification")

    @property
    def dae_residual_function(self):
        if hasattr(self, "_states_vector"):
            return self._expand_mx_func(
                ca.Function(
                    "dae_residual",
                    [
                        self.time,
                        self._states_vector,
                        self._der_states_vector,
                        self._alg_states_vector,
                        self._inputs_vector,
                        ca.veccat(*self._symbols(self.constants)),
                        ca.veccat(*self._symbols(self.parameters)),
                    ],
                    [ca.veccat(*self.equations)] if len(self.equations) > 0 else [],
                )
            )
        else:
            return self._expand_mx_func(
                ca.Function(
                    "dae_residual",
                    [
                        self.time,
                        ca.veccat(*self._symbols(self.states)),
                        ca.veccat(*self._symbols(self.der_states)),
                        ca.veccat(*self._symbols(self.alg_states)),
                        ca.veccat(*self._symbols(self.inputs)),
                        ca.veccat(*self._symbols(self.constants)),
                        ca.veccat(*self._symbols(self.parameters)),
                    ],
                    [ca.veccat(*self.equations)] if len(self.equations) > 0 else [],
                )
            )

    @property
    def initial_residual_function(self):
        if hasattr(self, "_states_vector"):
            return self._expand_mx_func(
                ca.Function(
                    "initial_residual",
                    [
                        self.time,
                        self._states_vector,
                        self._der_states_vector,
                        self._alg_states_vector,
                        self._inputs_vector,
                        ca.veccat(*self._symbols(self.constants)),
                        ca.veccat(*self._symbols(self.parameters)),
                    ],
                    [ca.veccat(*self.initial_equations)] if len(self.initial_equations) > 0 else [],
                )
            )
        else:
            return self._expand_mx_func(
                ca.Function(
                    "initial_residual",
                    [
                        self.time,
                        ca.veccat(*self._symbols(self.states)),
                        ca.veccat(*self._symbols(self.der_states)),
                        ca.veccat(*self._symbols(self.alg_states)),
                        ca.veccat(*self._symbols(self.inputs)),
                        ca.veccat(*self._symbols(self.constants)),
                        ca.veccat(*self._symbols(self.parameters)),
                    ],
                    [ca.veccat(*self.initial_equations)] if len(self.initial_equations) > 0 else [],
                )
            )

    @property
    def variable_metadata_function(self):
        in_var = ca.veccat(*self._symbols(self.parameters))
        out = []
        is_affine = True
        zero, one = ca.MX(0), ca.MX(1)  # Recycle these common nodes as much as possible.
        for variable_list in [
            self.states,
            self.alg_states,
            self.inputs,
            self.parameters,
            self.constants,
        ]:
            attribute_lists = [[] for i in range(len(CASADI_ATTRIBUTES))]
            for variable in variable_list:
                for attribute_list_index, attribute in enumerate(CASADI_ATTRIBUTES):
                    value = getattr(variable, attribute)
                    # Try casting to DM first, in case we have a nested list that needs to be interpreted as a matrix.
                    try:
                        value = ca.DM(value)
                    except Exception:
                        pass
                    value = ca.MX(value)
                    if value.is_zero():
                        value = zero
                    elif value.is_one():
                        value = one
                    value = (
                        value if value.numel() != 1 else ca.repmat(value, *variable.symbol.size())
                    )
                    attribute_lists[attribute_list_index].append(value)
            expr = ca.horzcat(*[ca.veccat(*attribute_list) for attribute_list in attribute_lists])
            if len(self.parameters) > 0 and isinstance(expr, ca.MX):
                f = ca.Function("f", [in_var], [expr])
                # NOTE: This is not a complete list of operations that can be
                # handled in an affine expression. That said, it should
                # capture the most common ways variable attributes are
                # expressed as a function of parameters.
                allowed_ops = {
                    ca.OP_INPUT,
                    ca.OP_OUTPUT,
                    ca.OP_CONST,
                    ca.OP_SUB,
                    ca.OP_ADD,
                    ca.OP_SUB,
                    ca.OP_MUL,
                    ca.OP_DIV,
                    ca.OP_NEG,
                }
                f_ops = {f.instruction_id(k) for k in range(f.n_instructions())}
                contains_unallowed_ops = not f_ops.issubset(allowed_ops)
                zero_hessian = ca.jacobian(ca.jacobian(expr, in_var), in_var).is_zero()
                if contains_unallowed_ops or not zero_hessian:
                    is_affine = False
            out.append(expr)
        if len(self.parameters) > 0 and is_affine:
            # Rebuild variable metadata as a single affine expression, if all
            # subexpressions are affine.
            in_var_ = ca.MX.sym("in_var", in_var.shape)
            out_ = []
            for o in out:
                Af = ca.Function("Af", [in_var], [ca.jacobian(o, in_var)])
                bf = ca.Function("bf", [in_var], [o])

                A = Af(0)
                A = ca.sparsify(A)

                b = bf(0)
                b = ca.sparsify(b)

                o_ = ca.reshape(ca.mtimes(A, in_var_), o.shape) + b
                out_.append(o_)
            out = out_
            in_var = in_var_

        return self._expand_mx_func(ca.Function("variable_metadata", [in_var], out))

    @property
    def delay_arguments_function(self):
        # We cannot assume that we can ca.horzcat/vertcat all delay arguments
        # and expressions due to shape differences, so instead we flatten our
        # delay expressions and durations into list, i.e. [delay_expr_1,
        # delay_duration_1, delay_expr_2, delay_duration_2, ...].

        out_arguments = list(itertools.chain.from_iterable(self.delay_arguments))

        if hasattr(self, "_states_vector"):
            return self._expand_mx_func(
                ca.Function(
                    "delay_arguments",
                    [
                        self.time,
                        self._states_vector,
                        self._der_states_vector,
                        self._alg_states_vector,
                        self._inputs_vector,
                        ca.veccat(*self._symbols(self.constants)),
                        ca.veccat(*self._symbols(self.parameters)),
                    ],
                    out_arguments,
                )
            )
        else:
            return self._expand_mx_func(
                ca.Function(
                    "delay_arguments",
                    [
                        self.time,
                        ca.veccat(*self._symbols(self.states)),
                        ca.veccat(*self._symbols(self.der_states)),
                        ca.veccat(*self._symbols(self.alg_states)),
                        ca.veccat(*self._symbols(self.inputs)),
                        ca.veccat(*self._symbols(self.constants)),
                        ca.veccat(*self._symbols(self.parameters)),
                    ],
                    out_arguments,
                )
            )



import copy
import os

from lxml import etree, objectify

from pymoca import ast
from pymoca.tree import TreeListener, TreeWalker, flatten

FILE_DIR = os.path.dirname(os.path.realpath(__file__))
BUILTINS = dir(__builtins__) + ["psi"]

E = objectify.E


class XmlGenerator(TreeListener):
    def __init__(self):
        super().__init__()
        self.xml = {}

    def exitEquation(self, tree: ast.Equation):
        self.xml[tree] = E(
            "equal",
            self.xml[tree.left],
            self.xml[tree.right],
        )

    def exitExpression(self, tree: ast.Expression):
        if isinstance(tree.operator, ast.ComponentRef):
            op_name = tree.operator.name
        else:
            op_name = tree.operator
        if len(tree.operands) == 1:
            self.xml[tree] = E("operator", *[self.xml[c] for c in tree.operands], name=op_name)
        else:
            self.xml[tree] = E("apply", *[self.xml[c] for c in tree.operands], builtin=op_name)

    def exitPrimary(self, tree: ast.Primary):
        self.xml[tree] = E("real", value=str(tree.value))

    def exitComponentRef(self, tree: ast.ComponentRef):
        self.xml[tree] = E("local", name=tree.name)

    def exitSymbol(self, tree: ast.Symbol):  # noqa: C901
        variability = None
        for v_type in ["discrete", "continuous", "parameter", "constant"]:
            if v_type in tree.prefixes:
                variability = v_type
                break
        items = []
        for f in ["start", "value"]:
            val = getattr(tree, f).value
            if val is None:
                continue
            items.append(E("item", E("real", value=str(val)), name=f))

        for f in ["fixed"]:
            val = getattr(tree, f).value
            if val is None:
                continue
            if val:
                val_xml = E("true")
            else:
                if f in ["fixed"]:
                    # this is already default
                    continue
                val_xml = E("false")
            items.append(E("item", val_xml, name=f))

        modifier = E(
            "modifier",
            *items,
        )
        e = E(
            "component",
            E("builtin", name=tree.type.name),
            modifier,
            name=tree.name,
        )  # type: etree._Element
        if variability is not None:
            e.attrib["variability"] = variability
        self.xml[tree] = e

    def exitClassModification(self, tree: ast.ClassModification):
        self.xml[tree] = E("modifier", *[self.xml[arg] for arg in tree.arguments])

    def exitFunction(self, tree: ast.Function):
        self.xml[tree] = E("apply", *[self.xml[arg] for arg in tree.arguments], builtin=tree.name)

    def exitWhenEquation(self, tree: ast.WhenEquation):
        self.xml[tree] = E(
            "when",
            E(
                "cond",
                self.xml[tree.conditions[0]],
            ),
            E(
                "then",
                *[self.xml[b] for b in tree.blocks[0]],
            ),
        )

    def exitClass(self, tree: ast.Class):
        self.xml[tree] = E(
            "classDefinition",
            E(
                "class",
                *[self.xml[s] for s in tree.symbols.values()],
                E("equation", *[self.xml[s] for s in tree.equations]),
                kind="model",
            ),
            name=tree.name,
        )

    def exitTree(self, tree: ast.Tree):
        self.xml[tree] = E(
            "modelica",
            E("declarations", *[self.xml[c] for c in tree.classes.values()]),
            format="1.0",
        )


def generate(ast_tree: ast.Tree, model_name: str):
    """
    :param ast_tree: AST to generate from
    :param model_name: class to generate
    :return: sympy source code for model
    """
    component_ref = ast.ComponentRef.from_string(model_name)
    ast_tree_new = copy.deepcopy(ast_tree)
    ast_walker = TreeWalker()
    flat_tree = flatten(ast_tree_new, component_ref)
    gen = XmlGenerator()
    ast_walker.walk(gen, flat_tree)
    return etree.tostring(gen.xml[flat_tree], pretty_print=True).decode("utf-8")



"""
This module contiains various mathematical models for a modelica model.
"""

from typing import Any, Dict, List, Union

import casadi as ca

SYM = Union[ca.SX, ca.MX]


class CasadiModel:
    def __init__(self, sym, *args, **kwargs):
        self.sym = sym

        # handle user args
        for k, v in kwargs.items():
            if k in self.__dict__.keys():
                setattr(self, k, v)
            else:
                raise ValueError("unknown argument", k)

        # Use just-in-time compilation to speed up the evaluation
        if ca.Importer.has_plugin("clang"):
            with_jit = True
            compiler = "clang"
        elif ca.Importer.has_plugin("shell"):
            with_jit = True
            compiler = "shell"
        else:
            print("WARNING; running without jit. " "This may result in very slow evaluation times")
            with_jit = False
            compiler = ""
        self.func_opt = {"jit": with_jit, "compiler": compiler}

    def __repr__(self):
        s = "\n"
        for k, v in sorted(self.__dict__.items()):
            if isinstance(v, self.sym):
                s += "{:8s}({:3d}):\t{:s}\n".format(k, v.shape[0], str(v))
        return s


class HybridOde(CasadiModel):
    """Hybrid Oridinary Differential Equation Model"""

    def __init__(self, sym: type = ca.SX, **kwargs):
        self.c = sym(0, 1)  # conditions
        self.dx = sym(0, 1)  # states derivatives
        self.f_c = sym(0, 1)  # condition relations
        self.f_i = sym(0, 1)  # reinit equations
        self.f_m = sym(0, 1)  # discrete update
        self.f_x_rhs = sym(0, 1)  # continuous integration
        self.y_rhs = sym(0, 1)  # algebraic states as a function of state
        self.m = sym(0, 1)  # discrete states
        self.ng = sym(0, 1)  # gaussian noise
        self.nu = sym(0, 1)  # uniform noise
        self.p = sym(0, 1)  # parameters and constants
        self.pre_m = sym(0, 1)  # discrete pre states
        self.pre_c = sym(0, 1)  # pre conditions
        self.prop = {}  # properties
        self.sym = sym  # symbol type
        self.t = sym()  # time
        self.x = sym(0, 1)  # states (have derivatives)
        self.y = sym(0, 1)  # algebraic states
        super().__init__(sym, **kwargs)

    def create_function_f_c(self):
        """condition function"""
        return ca.Function(
            "f_c",
            [self.t, self.x, self.y, self.m, self.p, self.ng, self.nu],
            [self.f_c],
            ["t", "x", "y", "m", "p", "ng", "nu"],
            ["c"],
            self.func_opt,
        )

    def create_function_f_i(self):
        """state reinitialization (reset) function"""
        return ca.Function(
            "f_i",
            [self.t, self.x, self.y, self.m, self.p, self.c, self.pre_c, self.ng, self.nu],
            [self.f_i],
            ["t", "x", "y", "m", "p", "c", "pre_c", "ng", "nu"],
            ["x_n"],
            self.func_opt,
        )

    def create_function_f_x_rhs(self):
        """ODE rhs for continuous state integration"""
        return ca.Function(
            "f_x_rhs",
            [self.t, self.x, self.y, self.m, self.p, self.c, self.ng, self.nu],
            [self.f_x_rhs],
            ["t", "x", "y", "m", "p", "c", "ng", "nu"],
            ["x_rhs"],
            self.func_opt,
        )

    def create_function_f_m(self):
        """Discrete state dynamics"""
        return ca.Function(
            "f_m",
            [self.t, self.x, self.y, self.m, self.p, self.c, self.pre_c, self.ng, self.nu],
            [self.f_m],
            ["t", "x", "y", "m", "p", "c", "pre_c", "ng", "nu"],
            ["m"],
            self.func_opt,
        )

    def create_function_f_J(self):
        """Jacobian for state integration"""
        return ca.Function(
            "J",
            [self.t, self.x, self.y, self.m, self.p, self.c, self.ng, self.nu],
            [ca.jacobian(self.f_x_rhs, self.x)],
            ["t", "x", "y", "m", "p", "c", "ng", "nu"],
            ["J"],
            self.func_opt,
        )

    def create_function_f_y(self):
        """output function"""
        return ca.Function(
            "y",
            [self.t, self.x, self.m, self.p, self.c, self.ng, self.nu],
            [self.y_rhs],
            ["t", "x", "m", "p", "c", "ng", "nu"],
            ["y"],
            self.func_opt,
        )


class HybridDae(CasadiModel):
    """Hybrid Differential Algebraic Equation Model"""

    def __init__(self, sym: type = ca.SX, **kwargs):
        self.sym = sym
        self.c = sym(0, 1)  # conditions
        self.dx = sym(0, 1)  # states derivatives
        self.f_c = sym(0, 1)  # condition relations
        self.f_i = sym(0, 1)  # reinit equations
        self.f_m = sym(0, 1)  # discrete update
        self.f_x = sym(0, 1)  # continuous integration
        self.m = sym(0, 1)  # discrete states
        self.ng = sym(0, 1)  # gaussian noise
        self.nu = sym(0, 1)  # uniform noise
        self.p = sym(0, 1)  # parameters and constants
        self.pre_m = sym(0, 1)  # discrete pre states
        self.pre_c = sym(0, 1)  # pre conditions
        self.prop = {}  # properties
        self.t = sym()  # time
        self.x = sym(0, 1)  # states (have derivatives)
        self.y = sym(0, 1)  # algebraic states

        super().__init__(sym, **kwargs)

    def to_ode(self) -> HybridOde:
        """Convert to a HybridOde"""
        res_split = split_dae_alg(self.f_x, self.dx)
        alg = res_split["alg"]
        dae = res_split["dae"]

        x_rhs = tangent_approx(dae, self.dx, assert_linear=True)
        y_rhs = tangent_approx(alg, self.y, assert_linear=True)

        return HybridOde(
            c=self.c,
            dx=self.dx,
            f_c=self.f_c,
            f_i=self.f_i,
            f_m=self.f_m,
            f_x_rhs=x_rhs,
            y_rhs=y_rhs,
            m=self.m,
            ng=self.ng,
            nu=self.nu,
            p=self.p,
            pre_m=self.pre_m,
            pre_c=self.pre_c,
            prop=self.prop,
            sym=self.sym,
            t=self.t,
            x=self.x,
            y=self.y,
        )


def split_dae_alg(eqs: SYM, dx: SYM) -> Dict[str, SYM]:
    """Split equations into differential algebraic and algebraic only"""
    dae = []
    alg = []
    for eq in ca.vertsplit(eqs):
        if ca.depends_on(eq, dx):
            dae.append(eq)
        else:
            alg.append(eq)
    return {"dae": ca.vertcat(*dae), "alg": ca.vertcat(*alg)}


def permute(x: SYM, perm: List[int]) -> SYM:
    """Perumute a vector"""
    x_s = []
    for i in perm:
        x_s.append(x[i])
    return ca.vertcat(*x_s)


def blt(f: List[SYM], x: List[SYM]) -> Dict[str, Any]:
    """
    Sort equations by dependence
    """
    J = ca.jacobian(f, x)
    nblock, rowperm, colperm, rowblock, colblock, coarserow, coarsecol = J.sparsity().btf()
    return {
        "J": J,
        "nblock": nblock,
        "rowperm": rowperm,
        "colperm": colperm,
        "rowblock": rowblock,
        "colblock": colblock,
        "coarserow": coarserow,
        "coarsecol": coarsecol,
    }


def tangent_approx(f: SYM, x: SYM, a: SYM = None, assert_linear: bool = False) -> Dict[str, SYM]:
    """
    Create a tangent approximation of a non-linear function f(x) about point a
    using a block lower triangular solver

    0 = f(x) = f(a) + J*x   # taylor series about a (if f(x) linear in x, then globally valid)
    J*x = -f(a)             # solve for x
    x = -J^{-1}f(a)         # but inverse is slow, so we use solve
    where J = df/dx
    """
    # find f(a)
    if a is None:
        a = ca.DM.zeros(x.numel(), 1)
    f_a = ca.substitute(f, x, a)  # f(a)
    J = ca.jacobian(f, x)
    if assert_linear and ca.depends_on(J, x):
        raise AssertionError("not linear")
    # solve is smart enough to to convert to blt if necessary
    return ca.solve(J, -f_a)



"""
Scipy based model simulation.
"""

import time
from typing import Dict

import casadi as ca

import numpy as np

import scipy.integrate

from .model import HybridOde


def sim(model: HybridOde, options: Dict = None, user_callback=None) -> Dict[str, np.array]:
    """
    Simulates a Dae model.

    :model: The model to simulate
    :options: See opt dict below
    :user_callback: A routine to call after each integration step,
      f(t, x, y, m, p, c) -> ret   (ret < 0 means abort)
    """
    if model.f_x_rhs.shape[0] < 1:
        raise ValueError(
            "there are no ODE equations to simulate, " "check that the model is explicit"
        )

    ic = {}
    for f in ["x", "y", "m", "p"]:
        ic[f] = []
        for x in ca.vertsplit(getattr(model, f)):
            start = model.prop[x.name()]["start"]
            value = model.prop[x.name()]["value"]
            if start is not None:
                ic[f].append(ca.reshape(start, x.numel(), 1))
            elif value is not None:
                ic[f].append(ca.reshape(value, x.numel(), 1))
            else:
                ic[f].append(ca.DM.zeros(x.numel(), 1))
                Warning("using default start value for", x.name())

        ic[f] = np.array([ic[f]], dtype=float).T

    # set options
    opt = {
        "x0": ic["x"],
        "p": ic["p"],
        "t0": 0,
        "tf": 1,
        "dt": None,
        "integrator": "dopri5",
        "atol": 1e-6,
        "rtol": 1e-6,
        "max_step": None,
        "record_event_times": True,
        "verbose": False,
    }
    if options is not None:
        for k in options.keys():
            if k in opt.keys():
                opt[k] = options[k]
            else:
                raise ValueError("unknown option {:s}".format(k))
    if opt["dt"] is None:
        opt["dt"] = opt["tf"] / 100
    if opt["max_step"] is None:
        opt["max_step"] = opt["dt"] / 2

    # create functions
    f_y = model.create_function_f_y()
    f_c = model.create_function_f_c()
    f_m = model.create_function_f_m()
    f_x_rhs = model.create_function_f_x_rhs()
    f_J = model.create_function_f_J()
    f_i = model.create_function_f_i()

    # initialize sim loop
    t0 = opt["t0"]
    tf = opt["tf"]
    x = opt["x0"]
    ng = np.zeros(model.ng.shape[0])
    nu = np.zeros(model.nu.shape[0])
    m = ic["m"]
    p = opt["p"]
    y0 = ic["y"]
    pre_c = np.array(f_c(t0, x, y0, m, p, ng, nu))
    c = pre_c
    y = f_y(t0, x, m, p, c, ng, nu)
    dt = opt["dt"]
    t_vect = np.arange(t0, tf, dt)
    n = len(t_vect)
    data = {
        "t": [],
        "x": [],
        "m": [],
        "y": [],
        "c": [],
    }

    # create integrator
    integrator = scipy.integrate.ode(f_x_rhs, f_J)
    integrator.set_integrator(
        opt["integrator"],
        first_step=opt["max_step"],
        atol=opt["atol"],
        rtol=opt["rtol"],
        max_step=opt["max_step"],
    )

    # try to catch events with sol out, (root finding)
    def sol_out(t, x):
        c = np.array(f_c(t, x, y, m, p, ng, nu))
        if np.any(c != pre_c):
            # print('event', t)
            return -1
        return 0

    if opt["integrator"] in ["dopri5", "dopri853"]:
        integrator.set_solout(sol_out)

    # run sim loop
    i = 0
    dt_f_i = 0
    dt_integrate = 0
    integrator.set_initial_value(opt["x0"], t0)

    while i < n:
        t = integrator.t

        # resample noise
        ng = np.random.randn(model.ng.shape[0])
        nu = np.random.randn(model.nu.shape[0])

        # call reinit
        start = time.time()
        x = f_i(t, x, y, m, p, c, pre_c, ng, nu)
        dt_f_i += time.time() - start

        # setup next continuous integration step
        integrator.set_initial_value(x, t)
        integrator.set_f_params(y, m, p, c, ng, nu)
        integrator.set_jac_params(y, m, p, c, ng, nu)

        # integrate
        t_goal = t0 + i * dt
        start = time.time()
        integrator.integrate(t_goal)
        dt_integrate += time.time() - start
        x = integrator.y

        # compute new conditions
        pre_c = c
        c = np.array(f_c(t, x, y, m, p, ng, nu))

        # compute output
        y = f_y(t, x, m, p, c, ng, nu)

        # compute discrete states
        m = f_m(t, x, y, m, p, c, pre_c, ng, nu)

        # store data
        if opt["record_event_times"] or (integrator.t - t_goal == 0):
            data["t"].append(t)
            data["x"].append(ca.vertsplit(x))
            data["y"].append(ca.vertsplit(y))
            data["m"].append(ca.vertsplit(m))
            data["c"].append(ca.vertsplit(c))
            if user_callback is not None:
                user_callback(t, x, y, m, p, c)

        # increment time goal if reached
        if integrator.t - t_goal == 0:
            # update discrete states
            # TODO: make this use sampling
            i += 1

    for k in data.keys():
        data[k] = np.array(data[k])

    data["labels"] = {}
    for field in ["x", "y", "c", "m"]:
        data["labels"][field] = [x.name() for x in ca.vertsplit(getattr(model, field))]

    if opt["verbose"]:
        print("dt_f_i\t\t\t:", dt_f_i)
        print("dt_integrate\t:", dt_integrate)
    return data
